import torch
import torch.nn as nn
SEloss = nn.MSELoss(reduction = 'sum')

class NMF_algorithim(nn.Module):
    ''' implements ARD NMF from https://arxiv.org/pdf/1111.6085.pdf '''
    def __init__(self,Beta,H_prior,W_prior):
        super(NMF_algorithim, self).__init__()
        # Beta paramaterizes the objective function
        # Beta = 1 induces a poisson objective
        # Beta = 2 induces a gaussian objective
        # Priors on the component matrices are Exponential (L1) and half-normal (L2)

        if Beta == 1 and H_prior == 'L1' and W_prior == 'L1' :
            self.update_W = update_W_poisson_L1
            self.update_H = update_H_poisson_L1
            self.lambda_update = update_lambda_L1

        elif Beta == 1 and H_prior == 'L1' and W_prior == 'L2':
            self.update_W = update_W_poisson_L2
            self.update_H = update_H_poisson_L1
            self.lambda_update = update_lambda_L2_L1

        elif Beta == 1 and H_prior == 'L2' and W_prior == 'L1':
            self.update_W = update_W_poisson_L1
            self.update_H = update_H_poisson_L2
            self.lambda_update = update_lambda_L1_L2

        elif Beta == 1 and H_prior == 'L2' and W_prior == 'L2':
            self.update_W = update_W_poisson_L2
            self.update_H = update_H_poisson_L2
            self.lambda_update = update_lambda_L2

        if Beta == 2 and H_prior == 'L1' and W_prior == 'L1':
            self.update_W = update_W_gaussian_L1
            self.update_H = update_H_gaussian_L1
            self.lambda_update = update_lambda_L1

        elif Beta == 2 and H_prior == 'L1' and W_prior == 'L2':
            self.update_W = update_W_gaussian_L2
            self.update_H = update_H_gaussian_L1
            self.lambda_update = update_lambda_L2_L1

        elif Beta == 2 and H_prior == 'L2' and W_prior == 'L1':
            self.update_W = update_W_gaussian_L1
            self.update_H = update_H_gaussian_L2
            self.lambda_update = update_lambda_L1_L2

        elif Beta == 2 and H_prior == 'L2' and W_prior == 'L2':
            self.update_W = update_W_gaussian_L2
            self.update_H = update_H_gaussian_L2
            self.lambda_update = update_lambda_L2

    def forward(self,W, H, V, lambda_, C, b0, eps_, phi):
        h_ = self.update_H(H, W, lambda_, phi, V, eps_)
        w_ = self.update_W(h_, W, lambda_, phi, V, eps_)
        lam_ = self.lambda_update(w_,h_,b0,C,eps_)
        return h_, w_,lam_



def beta_div(Beta,V,W,H,eps_):
    V_ap = torch.matmul(W, H).type(V.dtype) + eps_.type(V.dtype)
    if Beta == 2:
        return SEloss(V,V_ap)/2
    if Beta == 1:
        lr = torch.log(torch.div(V, V_ap))
        return torch.sum( ( (V*lr) + V_ap) - V)

def calculate_objective_function(Beta,V,W,H,lambda_,C, eps_,phi,K):
    loss = beta_div(Beta,V,W,H,eps_)
    cst = (K*C)*(1.0-torch.log(C))
    return torch.pow(phi,-1)*loss + (C*torch.sum(torch.log(lambda_ * C))) + cst

def update_H_poisson_L1(H, W, lambda_, phi, V, eps_):
    #beta = 1 gamma(beta) = 1
    denom = torch.sum(W, 0) + torch.div(phi, lambda_) + eps_
    V_ap = torch.matmul(W, H) + eps_
    V_res = torch.div(V, V_ap)
    update = torch.div(torch.matmul(W.transpose(1,0), V_res), denom.reshape(-1,1))
    return H * update

def update_H_poisson_L2(H,W,lambda_,phi,V, eps_):
    #beta = 1 zeta(beta) = 1/2
    denom = torch.sum(W,0).reshape(-1,1) + torch.div(phi*H, lambda_.reshape(-1,1)) + eps_
    V_ap = torch.matmul(W, H) + eps_
    update = torch.pow(torch.div(torch.matmul(W.transpose(0,1), torch.div(V, V_ap)), denom),0.5)
    return H * update

def update_H_gaussian_L1(H,W,lambda_,phi,V,eps_):
    #beta = 2 gamma(beta) = 1
    V_ap = torch.matmul(W, H) + eps_
    denom = torch.matmul(W.transpose(0,1),V_ap) + torch.div(phi, lambda_ ).reshape(-1,1) + eps_
    update = torch.div(torch.matmul(W.transpose(0,1),V),denom)
    return H * update

def update_H_gaussian_L2(H,W,lambda_,phi,V,eps_):
    #beta = 2 zeta(beta) = 1
    denom = torch.matmul(W.transpose(0,1).type(V.dtype),torch.matmul(W, H).type(V.dtype) + eps_) + torch.div(phi * H, lambda_.reshape(-1,1)).type(V.dtype) + eps_
    update = torch.div(torch.matmul(W.transpose(0,1).type(V.dtype),V),denom)
    return H * update.type(torch.float32)

def update_W_poisson_L1(H, W, lambda_, phi, V, eps_):
    #beta = 1 gamma(beta) = 1
    denom = torch.sum(H, 1) + torch.div(phi, lambda_ ) + eps_
    V_ap = torch.matmul(W, H) + eps_
    V_res = torch.div(V, V_ap)
    update = torch.div(torch.matmul(V_res, H.transpose(0,1)), denom)
    return W * update

def update_W_poisson_L2(H,W,lambda_,phi,V,eps_):
    # beta = 1 zeta(beta) = 1/2
    V_ap = torch.matmul(W,H) + eps_
    V_res = torch.div(V, V_ap)
    denom = torch.sum(H,1) + torch.div(phi*W,lambda_) + eps_
    update = torch.pow(torch.div(torch.matmul(V_res,H.transpose(0,1)),denom),0.5)
    return W * update

def update_W_gaussian_L1(H,W,lambda_,phi,V,eps_):
    #beta = 2 gamma(beta) = 1
    V_ap = torch.matmul(W,H).type(V.dtype) + eps_
    denom = torch.matmul(V_ap,H.transpose(0,1).type(V.dtype)) + torch.div(phi,lambda_).type(V.dtype) + eps_
    update = torch.div(torch.matmul(V,H.transpose(0,1).type(V.dtype)),denom)
    return W * update.type(torch.float32)

def update_W_gaussian_L2(H,W,lambda_,phi,V,eps_):
    #beta = 2 zeta(beta) = 1
    V_ap = torch.matmul(W,H) + eps_
    denom = torch.matmul(V_ap,H.transpose(0,1)) + torch.div(phi*W,lambda_) + eps_
    update = torch.div(torch.matmul(V,H.transpose(0,1)),denom)
    return W * update

# update tolerance value for early stop criteria
def update_del(lambda_, lambda_last):
    del_ = torch.max(torch.div(torch.abs(lambda_ - lambda_last)), lambda_last)
    return del_



def update_lambda_L1(W,H,b0,C,eps_):
    return torch.div(torch.sum(W,0) + torch.sum(H,1) + b0, C)

def update_lambda_L2(W,H,b0,C,eps_):
    return torch.div(0.5*torch.sum(W*W,0) + (0.5*torch.sum(H*H,1))+b0,C)

def update_lambda_L1_L2(W,H,b0,C,eps_):
    return torch.div(torch.sum(W,0) + 0.5*torch.sum(H*H,1)+b0,C)

def update_lambda_L2_L1(W,H,b0,C,eps_):
    return torch.div(0.5*torch.sum(torch.pow(W,2),0) + torch.sum(H,1)+b0,C)



import sys
import argparse
import os
import pkg_resources
import pandas as pd
from typing import Union
import numpy as np
import matplotlib.pyplot as plt

from .utils import postprocess_msigs, get_nlogs_from_output, file_loader
from .utils import load_cosmic_signatures
from .utils import split_negatives

from .consensus import consensus_cluster

from .plotting import k_dist, consensus_matrix
from .plotting import signature_barplot, stacked_bar, signature_barplot_DBS
from .plotting import marker_heatmap
from .plotting import cosine_similarity_plot

from .spectra import get_spectra_from_maf
from .bnmf import ardnmf

def run_maf(
    maf: Union[str, pd.DataFrame],
    outdir: str = '.',
    cosmic: str = 'cosmic2',
    hg_build: Union[str, None] = None,
    nruns: int = 10,
    verbose: bool = False,
    plot_results: bool = True,
    **nmf_kwargs
    ):
    """
    Args:
        * maf: input .maf file format
        * outdir: output directory to save files
        * cosmic: cosmic signature set to use
        * hg_build: human genome build for generating reference context
        * nruns: number of iterations for ARD-NMF
        * verbose: bool

    NMF_kwargs:
        * K0: starting number of latent components
        * objective: objective function for optimizaiton
        * max_iter: maximum number of iterations for algorithm
        * del_: n/a
        * tolerance: stop point for optimization
        * phi: dispersion parameter
        * a: shape parameter
        * b: shape parameter
        * prior_on_W: L1 or L2
        * prior_on_H: L1 or L2
        * report_freq: how often to print stats
        * active_thresh: threshold for a latent component's impact on
            signature if the latent factor is less than this, it does not contribute
        * cut_norm: min normalized value for mean signature
            (used in post-processing)
        * cut_diff: difference between mean signature and rest of signatures
            for marker selction
            (used in post-processing)
        * cuda_int: GPU to use. Defaults to 0. If "None" or if no GPU available,
            will perform decomposition using CPU.
    """
    try:
        [nmf_kwargs.pop(key) for key in ['input', 'type']]
    except:
        pass

    if outdir is not ".":
        print("   * Creating output dir at {}".format(outdir))
        os.makedirs(outdir, exist_ok=True)

    # Human Genome Build
    if hg_build is not None:
        print("   * Using {} build".format(hg_build.split("/")[-1].split('.2bit')[0]))

    # Cosmic Signatures
    cosmic_df, cosmic_index = load_cosmic_signatures(cosmic)

    # Generate Spectra from Maf
    print("   * Loading spectra from {}".format(maf))
    maf, spectra = get_spectra_from_maf(
        pd.read_csv(maf, sep='\t'),
        hgfile=hg_build,
        cosmic=cosmic
    )

    print("   * Saving ARD-NMF outputs to {}".format(os.path.join(outdir,'nmf_output.h5')))
    store = pd.HDFStore(os.path.join(outdir,'nmf_output.h5'),'w')

    print("   * Running ARD-NMF...")
    for n_iter in range(nruns):
        store['X'] = spectra

        res = ardnmf(
            spectra,
            tag="\t{}/{}: ".format(n_iter,nruns-1),
            verbose=verbose,
            **nmf_kwargs
        )

        postprocess_msigs(res, cosmic_df, cosmic_index, cosmic)
        lam = pd.DataFrame(data=res["lam"], columns=["lam"])
        lam.index.name = "K0"

        store["run{}/H".format(n_iter)] = res["H"]
        store["run{}/W".format(n_iter)] = res["W"]
        store["run{}/lam".format(n_iter)] = lam
        store["run{}/Hraw".format(n_iter)] = res["Hraw"]
        store["run{}/Wraw".format(n_iter)] = res["Wraw"]
        store["run{}/markers".format(n_iter)] = res["markers"]
        store["run{}/signatures".format(n_iter)] = res["signatures"]
        store["run{}/log".format(n_iter)] = res["log"]
        store["run{}/cosine".format(n_iter)] = res["cosine"]

    store.close()

    # Select Best Result
    aggr = get_nlogs_from_output(os.path.join(outdir,'nmf_output.h5'))
    max_k = aggr.groupby("K").size().idxmax()
    max_k_iter = aggr[aggr['K']==max_k].shape[0]
    best_run = int(aggr[aggr['K']==max_k].obj.idxmin())
    print("   * Run {} had lowest objective with mode (n={:g}) K = {:g}.".format(best_run, max_k_iter, aggr.loc[best_run]['K']))

    store = pd.HDFStore(os.path.join(outdir,'nmf_output.h5'),'a')
    store["H"] = store["run{}/H".format(best_run)]
    store["W"] = store["run{}/W".format(best_run)]
    store["lam"] = store["run{}/lam".format(best_run)]
    store["Hraw"] = store["run{}/Hraw".format(best_run)]
    store["Wraw"] = store["run{}/Wraw".format(best_run)]
    store["markers"] = store["run{}/markers".format(best_run)]
    store["signatures"] = store["run{}/signatures".format(best_run)]
    store["log"] = store["run{}/log".format(best_run)]
    store["cosine"] = store["run{}/cosine".format(best_run)]
    store["aggr"] = aggr
    store.close()

    # Plots
    if plot_results:
        print("   * Saving report plots to {}".format(outdir))
        H = pd.read_hdf(os.path.join(outdir,'nmf_output.h5'), "H")
        W = pd.read_hdf(os.path.join(outdir,'nmf_output.h5'), "W")
        cosine = pd.read_hdf(os.path.join(outdir,'nmf_output.h5'), "cosine")

        if cosmic == 'cosmic3_DBS':
            _ = signature_barplot_DBS(W, contributions=np.sum(H))
        else:
            _ = signature_barplot(W, contributions=np.sum(H))

        plt.savefig(os.path.join(outdir, "signature_contributions.pdf"), dpi=100, bbox_inches='tight')
        _ = stacked_bar(H)
        plt.savefig(os.path.join(outdir, "signature_stacked_barplot.pdf"), dpi=100, bbox_inches='tight')
        _ = k_dist(np.array(aggr.K, dtype=int))
        plt.savefig(os.path.join(outdir, "k_dist.pdf"), dpi=100, bbox_inches='tight')
        _ = cosine_similarity_plot(cosine)
        plt.savefig(os.path.join(outdir, "cosine_similarity_plot.pdf"), dpi=100, bbox_inches='tight')

def run_spectra(
    spectra: Union[str, pd.DataFrame],
    outdir: str = '.',
    cosmic: str = 'cosmic2',
    nruns: int = 10,
    verbose: bool = False,
    plot_results: bool = True,
    **nmf_kwargs
    ):
    """
    Args:
        * spectra: filepath or pd.DataFrame of input spectra file (context x samples)
            NOTE: index should be context in the following format (1234): 3[1>2]4
        * outdir: output directory to save files
        * cosmic: cosmic signature set to use
        * nruns: number of iterations for ARD-NMF
        * verbose: bool

    NMF_kwargs:
        * K0: starting number of latent components
        * objective: objective function for optimizaiton
        * max_iter: maximum number of iterations for algorithm
        * del_: n/a
        * tolerance: stop point for optimization
        * phi: dispersion parameter
        * a: shape parameter
        * b: shape parameter
        * prior_on_W: L1 or L2
        * prior_on_H: L1 or L2
        * report_freq: how often to print stats
        * active_thresh: threshold for a latent component's impact on
            signature if the latent factor is less than this, it does not contribute
        * cut_norm: min normalized value for mean signature
            (used in post-processing)
        * cut_diff: difference between mean signature and rest of signatures
            for marker selction
            (used in post-processing)
        * cuda_int: GPU to use. Defaults to 0. If "None" or if no GPU available,
            will perform decomposition using CPU.
    """
    try:
        [nmf_kwargs.pop(key) for key in ['input', 'type', 'hg_build']]
    except:
        pass

    # Load spectra
    if isinstance(spectra, str):
        spectra = file_loader(spectra)

    if outdir is not ".":
        print("   * Creating output dir at {}".format(outdir))
        os.makedirs(outdir, exist_ok=True)

    # Cosmic Signatures
    cosmic_df, cosmic_index = load_cosmic_signatures(cosmic)

    print("   * Saving ARD-NMF outputs to {}".format(os.path.join(outdir,'nmf_output.h5')))
    store = pd.HDFStore(os.path.join(outdir,'nmf_output.h5'),'w')

    print("   * Running ARD-NMF...")
    for n_iter in range(nruns):
        store['X'] = spectra

        res = ardnmf(
            spectra,
            tag="\t{}/{}: ".format(n_iter,nruns-1),
            verbose=verbose,
            **nmf_kwargs
        )

        postprocess_msigs(res, cosmic_df, cosmic_index, cosmic)
        lam = pd.DataFrame(data=res["lam"], columns=["lam"])
        lam.index.name = "K0"

        store["run{}/H".format(n_iter)] = res["H"]
        store["run{}/W".format(n_iter)] = res["W"]
        store["run{}/lam".format(n_iter)] = lam
        store["run{}/Hraw".format(n_iter)] = res["Hraw"]
        store["run{}/Wraw".format(n_iter)] = res["Wraw"]
        store["run{}/markers".format(n_iter)] = res["markers"]
        store["run{}/signatures".format(n_iter)] = res["signatures"]
        store["run{}/log".format(n_iter)] = res["log"]
        store["run{}/cosine".format(n_iter)] = res["cosine"]

    store.close()

    # Select Best Result
    aggr = get_nlogs_from_output(os.path.join(outdir,'nmf_output.h5'))
    max_k = aggr.groupby("K").size().idxmax()
    max_k_iter = aggr[aggr['K']==max_k].shape[0]
    best_run = int(aggr[aggr['K']==max_k].obj.idxmin())
    print("   * Run {} had lowest objective with mode (n={:g}) K = {:g}.".format(best_run, max_k_iter, aggr.loc[best_run]['K']))

    store = pd.HDFStore(os.path.join(outdir,'nmf_output.h5'),'a')
    store["H"] = store["run{}/H".format(best_run)]
    store["W"] = store["run{}/W".format(best_run)]
    store["lam"] = store["run{}/lam".format(best_run)]
    store["Hraw"] = store["run{}/Hraw".format(best_run)]
    store["Wraw"] = store["run{}/Wraw".format(best_run)]
    store["markers"] = store["run{}/markers".format(best_run)]
    store["signatures"] = store["run{}/signatures".format(best_run)]
    store["log"] = store["run{}/log".format(best_run)]
    store["cosine"] = store["run{}/cosine".format(best_run)]
    store["aggr"] = aggr
    store.close()

    # Plots
    if plot_results:
        print("   * Saving report plots to {}".format(outdir))
        H = pd.read_hdf(os.path.join(outdir,'nmf_output.h5'), "H")
        W = pd.read_hdf(os.path.join(outdir,'nmf_output.h5'), "W")
        cosine = pd.read_hdf(os.path.join(outdir,'nmf_output.h5'), "cosine")

        if cosmic == 'cosmic3_DBS':
            _ = signature_barplot_DBS(W, contributions=np.sum(H))
        else:
            _ = signature_barplot(W, contributions=np.sum(H))

        plt.savefig(os.path.join(outdir, "signature_contributions.pdf"), dpi=100, bbox_inches='tight')
        _ = stacked_bar(H)
        plt.savefig(os.path.join(outdir, "signature_stacked_barplot.pdf"), dpi=100, bbox_inches='tight')
        _ = k_dist(np.array(aggr.K, dtype=int))
        plt.savefig(os.path.join(outdir, "k_dist.pdf"), dpi=100, bbox_inches='tight')
        _ = cosine_similarity_plot(cosine)
        plt.savefig(os.path.join(outdir, "cosine_similarity_plot.pdf"), dpi=100, bbox_inches='tight')

def run_matrix(
    matrix: Union[str, pd.DataFrame],
    outdir: str = '.',
    nruns: int = 20,
    verbose: bool = False,
    plot_results: bool = True,
    **nmf_kwargs
    ):
    """
    Args:
        * matrix: expression matrix; this should be normalized to accomodate
            Gaussian noise assumption (log2-norm) (n_features x n_samples)

            NOTE: recommended to filter out lowly expressed genes for RNA:
            *************** example filtering ***************
            tpm = tpm[
                (np.sum(tpm >= 0.1, 1) > tpm.shape[1]*0.2) &
                (np.sum(counts.iloc[:,1:] >= 6, 1) > tpm.shape[1]*0.2)
            ]
            *************************************************

            NOTE: reccomended to select a set of highly variable genes following
                this (~ 2000 - 7500 genes)

        * outdir: output directory to save files
        * cosmic: cosmic signature set to use
        * nruns: number of iterations for ARD-NMF
        * verbose: bool

    NMF_kwargs:
        * K0: starting number of latent components
        * objective: objective function for optimizaiton
        * max_iter: maximum number of iterations for algorithm
        * del_: n/a
        * tolerance: stop point for optimization
        * phi: dispersion parameter
        * a: shape parameter
        * b: shape parameter
        * prior_on_W: L1 or L2
        * prior_on_H: L1 or L2
        * report_freq: how often to print stats
        * active_thresh: threshold for a latent component's impact on
            signature if the latent factor is less than this, it does not contribute
        * cut_norm: min normalized value for mean signature
            (used in post-processing)
        * cut_diff: difference between mean signature and rest of signatures
            for marker selction
            (used in post-processing)
        * cuda_int: GPU to use. Defaults to 0. If "None" or if no GPU available,
            will perform decomposition using CPU.
    """
    try:
        [nmf_kwargs.pop(key) for key in ['input', 'type', 'hg_build', 'cosmic']]
    except:
        pass

    # Load matrix
    if isinstance(matrix, str):
        matrix = file_loader(matrix)

    # Check for negativity
    if min(matrix.min()) < 0:
        print("   * Negative values detecting, splitting vars m={} --> m={}".format(matrix.shape[0], matrix.shape[0]*2))
        matrix = split_negatives(matrix, axis=0)

    if outdir is not ".":
        print("   * Creating output dir at {}".format(outdir))
        os.makedirs(outdir, exist_ok=True)

    print("   * Saving ARD-NMF outputs to {}".format(os.path.join(outdir,'nmf_output.h5')))
    store = pd.HDFStore(os.path.join(outdir,'nmf_output.h5'),'w')

    print("   * Running ARD-NMF...")
    for n_iter in range(nruns):
        store['X'] = matrix

        res = ardnmf(
            matrix,
            tag="\t{}/{}: ".format(n_iter,nruns-1),
            verbose=verbose,
            **nmf_kwargs
        )

        lam = pd.DataFrame(data=res["lam"], columns=["lam"])
        lam.index.name = "K0"

        store["run{}/H".format(n_iter)] = res["H"]
        store["run{}/W".format(n_iter)] = res["W"]
        store["run{}/lam".format(n_iter)] = lam
        store["run{}/Hraw".format(n_iter)] = res["Hraw"]
        store["run{}/Wraw".format(n_iter)] = res["Wraw"]
        store["run{}/markers".format(n_iter)] = res["markers"]
        store["run{}/signatures".format(n_iter)] = res["signatures"]
        store["run{}/log".format(n_iter)] = res["log"]

    store.close()

    # Select Best Result
    aggr = get_nlogs_from_output(os.path.join(outdir,'nmf_output.h5'))
    max_k = aggr.groupby("K").size().idxmax()
    max_k_iter = aggr[aggr['K']==max_k].shape[0]
    best_run = int(aggr[aggr['K']==max_k].obj.idxmin())
    print("   * Run {} had lowest objective with mode (n={:g}) K = {:g}.".format(best_run, max_k_iter, aggr.loc[best_run]['K']))

    store = pd.HDFStore(os.path.join(outdir,'nmf_output.h5'),'a')
    store["H"] = store["run{}/H".format(best_run)]
    store["W"] = store["run{}/W".format(best_run)]
    store["lam"] = store["run{}/lam".format(best_run)]
    store["Hraw"] = store["run{}/Hraw".format(best_run)]
    store["Wraw"] = store["run{}/Wraw".format(best_run)]
    store["markers"] = store["run{}/markers".format(best_run)]
    store["signatures"] = store["run{}/signatures".format(best_run)]
    store["log"] = store["run{}/log".format(best_run)]
    store["aggr"] = aggr
    store.close()


    # Consensus Clustering
    print("   * Computing consensus matrix")
    cmatrix, _ = consensus_cluster(os.path.join(outdir, 'nmf_output.h5'))
    f,d = consensus_matrix(cmatrix, n_clusters=max_k_iter)

    cmatrix.to_csv(os.path.join(outdir, 'consensus_matrix.tsv'), sep='\t')
    d.to_csv(os.path.join(outdir, 'consensus_assign.tsv'), sep='\t')

    if plot_results: plt.savefig(os.path.join(outdir, 'consensus_matrix.pdf'), dpi=100, bbox_inches='tight')

    store = pd.HDFStore(os.path.join(outdir,'nmf_output.h5'),'a')
    store['consensus'] = d
    store.close()

    # Plots
    if plot_results:
        print("   * Saving report plots to {}".format(outdir))
        H = pd.read_hdf(os.path.join(outdir,'nmf_output.h5'), "H")
        X = pd.read_hdf(os.path.join(outdir,'nmf_output.h5'), "X")
        signatures = pd.read_hdf(os.path.join(outdir,'nmf_output.h5'), "signatures")

        _ = k_dist(np.array(aggr.K, dtype=int))
        plt.savefig(os.path.join(outdir, "k_dist.pdf"), dpi=100, bbox_inches='tight')

        _ = marker_heatmap(X, signatures, H.sort_values('max_id').max_id)
        plt.savefig(os.path.join(outdir, "marker_heatmap.pdf"), dpi=100, bbox_inches='tight')



import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
from typing import Union
import numpy as np
import pandas as pd

from ._utils import series_to_colors
from ._utils import color_list_to_matrix_and_cmap

def marker_heatmap(
    X: pd.DataFrame,
    signatures: pd.DataFrame,
    order_series: pd.Series,
    subset_genes: Union[pd.Series,None] = None,
    diff: float = 0.5,
    max_norm: float = 0.5,
    figsize: tuple = (16,12),
    cmap: str ="YlGnBu",
    display_y: bool = False,
    vmax: float = None,
    vmin: float = None,
    cohort_s: Union[pd.Series,None] = None,
    y_hm_label: str = 'Genes',
    cbar_hm_label: str = 'Normalized Expression'
    ):
    """
    Plot marker map.
    -----------------------------
    Args:
        * X: pd.DataFrame of input sample x feature matrix
        * signatures: pd.DataFrame signatures output;
            this bundles information about the weightings of each feature (ex. gene) and
            what signature they map to
        * order_series: series of samples mapping to subgroups
            index: X.index
            values: subgrouping
        * subset_series: a pd.Series with the index as the gene name or ID that
            matches the marker matrix & has a "Subgroup" column for labeling
        * diff: difference of loading for called signature vs. rest
        * max_norm: strength of loading for called signature
        * figsize: size of figure
        * cmap: colormap for plot
        * display_y: whether or not to display feature names
        * vmax: colorbar max
        * vmin: colorbar min
        * cohort_s: cohort_series dataframe (added on top of plot)
        * y_hm_label: label of y-axis on heatmap (ex. Genes, Protein LFC, etc.)
        * cbar_hm_label: label of heatmap colorbar

    Returns:
        * plt.Figure
    """
    # Filter for marker genes
    signatures_filt = signatures[(signatures['diff'] > diff) & (signatures['max_norm'] > max_norm)]

    # Remove signatures with no marker genes associated
    order_series = order_series[order_series.isin(set(signatures_filt['max_id'].astype(int)))]

    # Filter X matrix
    sample_markers = X.loc[signatures_filt.index, order_series.sort_values().index]

    # Set horizontal lines
    hz_lines = np.unique(sample_markers.join(signatures_filt).loc[:,'max_id'].values, return_index=True)[1]

    fig, ax = plt.subplots(figsize=figsize)

    x0 = ax.get_position().x0
    x1 = ax.get_position().x1
    y0 = ax.get_position().y0
    y1 = ax.get_position().y1
    buf = y1*0.01

    cbar_ax = fig.add_axes([.91, 0.5, .025, .3])

    sns.heatmap(sample_markers, ax=ax, cmap=cmap, rasterized=True, cbar_ax=cbar_ax, vmax=vmax, vmin=vmin)
    v,c = np.unique(order_series, return_counts=True)

    # plot horizontal lines
    _c = np.cumsum(c)
    _ci = np.roll(_c,2)
    _ci[0] = 0
    _ci[1] = 0
    ax.hlines(hz_lines, _ci, _c, rasterized=True)

    # plot vertical lines
    _h = list(hz_lines)
    _h.append(sample_markers.shape[0])
    ax.vlines(np.cumsum(c)[:-1], _h[:-2], _h[2:], rasterized=True)
    ax.vlines(np.cumsum(c)[:-1], 0, sample_markers.shape[0], alpha=0.4, rasterized=True)

    # set ticks
    ax.set_xticks(np.cumsum(c)-c/2)
    ax.set_xticklabels(v, rotation=360,fontsize=14)
    ax.set_yticks(np.arange(sample_markers.index.values.shape[0]))

    # add gene markings
    if subset_genes is not None:
        ax.set_yticks([])
        ax.set_yticklabels([], rasterized=True)

        lax = fig.add_axes([x0-3*buf, y0, 2*buf, y1-y0])
        lax.set_xticks([])
        lax.set_yticks([])

        meta = sample_markers.drop(columns=list(sample_markers)).join(subset_genes).iloc[:,0]

        colors_conversion, meta_colormap = series_to_colors(meta)
        meta_colormap_inv = dict([[v,k] for k,v in meta_colormap.items()])
        meta_colormap_inv = {(k[0],k[1],k[2]):v for k,v in meta_colormap_inv.items()}
        cbar_lax = fig.add_axes([0.06, 0.68, 2*buf, .2])

        # Add heatmapping
        mat,cmap = color_list_to_matrix_and_cmap(colors_conversion)
        sns.heatmap(
            mat.T,
            cmap=cmap,
            ax=lax,
            yticklabels=False,
            xticklabels=False,
            cbar=True,
            cbar_ax=cbar_lax,
        )

        cb_ticks = [float(t.get_text().replace('−','-')) for t in cbar_lax.get_yticklabels()]

        color_value_mapping = dict()
        for v in np.unique(mat):
            color_code = list(cmap.__call__(v))
            color_code = tuple(color_code[:3])
            color_value_mapping[v] = meta_colormap_inv[color_code]

        cbar_lax.get_yaxis().set_ticks([])

        n_labels = len(list(color_value_mapping.keys()))
        vals = [x * ((n_labels)/(n_labels+1)) + 0.5 * ((n_labels)/(n_labels+1)) for x in list(color_value_mapping.keys())]

        cbar_lax.get_yaxis().set_ticks(vals)
        cbar_lax.get_yaxis().set_ticklabels(list(color_value_mapping.values()),)
        cbar_lax.yaxis.set_ticks_position('left')

        cbar_lax.set_frame_on(True)

        lax.set_ylabel('Marker Genes', fontsize=14)
        ax.set_ylabel("")

        for _, spine in lax.spines.items():
            spine.set_visible(True)
    # --------------subset genes-------------------
    else:
        if display_y:
            ax.set_yticklabels(sample_markers.index.values, fontsize=5, rasterized=True)
        else:
            ax.set_yticks([])
            ax.set_yticklabels([], rasterized=True)

        ax.set_ylabel(y_hm_label, fontsize=14)

    # Add sample anont
    if cohort_s is not None:
        # Get ordering and samples
        cohort_s = cohort_s.loc[sample_markers.columns]

        # Create axis
        cs_ax = fig.add_axes([x0, y1+2*buf, x1*.86, 2*buf])
        cs_ax.set_xticks([])
        cs_ax.set_yticks([])

        cbar_cs_ax = fig.add_axes([x0, y1+6*buf, x1*.25, 2*buf])

        colors_conversion, meta_colormap = series_to_colors(cohort_s)
        meta_colormap_inv = dict([[v,k] for k,v in meta_colormap.items()])
        meta_colormap_inv = {(k[0],k[1],k[2]):v for k,v in meta_colormap_inv.items()}
        mat,cmap = color_list_to_matrix_and_cmap(colors_conversion)

        sns.heatmap(
            mat,
            cmap=cmap,
            ax=cs_ax,
            yticklabels=False,
            xticklabels=False,
            cbar=True,
            cbar_ax=cbar_cs_ax,
            cbar_kws={"orientation": "horizontal"}
        )

        cb_ticks = [float(t.get_text().replace('−','-')) for t in cbar_cs_ax.get_yticklabels()]

        color_value_mapping = dict()
        for v in np.unique(mat):
            color_code = list(cmap.__call__(v))
            color_code = tuple(color_code[:3])
            color_value_mapping[v] = meta_colormap_inv[color_code]

        cbar_cs_ax.get_xaxis().set_ticks([])

        n_labels = len(list(color_value_mapping.keys()))
        vals = [x * ((n_labels)/(n_labels+1)) + 0.5 * ((n_labels)/(n_labels+1)) for x in list(color_value_mapping.keys())]

        cbar_cs_ax.get_xaxis().set_ticks(vals)
        cbar_cs_ax.get_xaxis().set_ticklabels(list(color_value_mapping.values()))
        cbar_cs_ax.xaxis.set_ticks_position('top')

        cbar_cs_ax.set_frame_on(True)
        [spine.set_visible(True) for _, spine in cs_ax.spines.items()]
    # --------------sample annot-------------------

    ax.set_title('')
    ax.set_xlabel('NMF Signatures', fontsize=14)

    cbar_ax.set_ylabel(cbar_hm_label, fontsize=12)

    [spine.set_visible(True) for _, spine in ax.spines.items()]

    return fig



import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
from typing import Union
import numpy as np
import matplotlib as mpl
from matplotlib.pyplot import cm

from scipy.cluster import hierarchy
import scipy.cluster.hierarchy as shc
from sklearn.cluster import AgglomerativeClustering

from ._utils import series_to_colors
from ._utils import color_list_to_matrix_and_cmap

def k_dist(X: np.ndarray, figsize: tuple = (8,8)):
    """
    Selected signatures plot.
    --------------------------------------
    Args:
        * X: numbers to plot
        * figsize: size of figure (int,int)

    Returns:
        * fig

    Example usage:
        plot_k_dist(np.array(pd.read_hdf("nmf_output.h5","aggr").K))

    """
    fig,ax = plt.subplots(figsize=figsize)

    sns.countplot(X, ax=ax, linewidth=2, edgecolor='k', rasterized=True)
    ax.set_ylim(0,ax.get_ylim()[1]+int(ax.get_ylim()[1]*0.1))

    ax.set_title("Aggregate of ARD-NMF (n={})".format(len(X)), fontsize=20)
    ax.set_ylabel("Number of Iterations", fontsize=18)
    ax.set_xticklabels(ax.get_xticklabels(), fontsize=18)

    return fig


def consensus_matrix(
    cmatrix: pd.DataFrame,
    metric: str = 'euclidean',
    method: str = 'ward',
    n_clusters: int = 10,
    color_thresh_scale: float = 0.3,
    figsize: tuple = (8,8),
    p: int = 30,
    metas: Union[list, None] = None
    ):
    """
    Plot consensus matrix.
    -----------------------
    Args:
        * cmatrix: consensus matrix. This may be generated by calling:
            df, assign_p = consensus_cluster_ardnmf(filepath)
        * metric: distance metric
        * method: method of clustering
        * n_clusters: number of clusters for agglomerative clustering
        * color_thresh_scale: asthetic scale for coloring of dendrogram
        * figsize: figsize
        * p: parameter for dendrogram
        * meta: list of pd.Series that includes a variable of interest to plot
            to left of plot; must be categorical in nature

    Returns:
        * fig
    """
    # -------------
    # Heatmap
    # -------------
    fig,ax = plt.subplots(figsize=figsize)
    cbar_ax = fig.add_axes([ax.get_position().x1 + ax.get_position().x1*0.1, ax.get_position().y0, .025, .1])

    # Compute initial linkage to grab ordering
    d_linkage = shc.linkage(cmatrix, metric=metric, method=method)
    dres = shc.dendrogram(d_linkage, p=p, no_plot=True)
    dgram_idx = list(map(int, dres['ivl']))

    # Create heatmap
    sns.heatmap(
        cmatrix.iloc[dgram_idx,dgram_idx].values,
        ax=ax,
        square=True,
        cbar_ax=cbar_ax,
        cbar_kws = {'ticks':[0, np.max(cmatrix.values)]},
        rasterized=True
    )

    cbar_ax.set_ylabel('ARD-NMF \nMembership', fontsize=10,rotation=90)
    ax.set_xticks([])
    ax.set_yticks([])

    # -------------
    # Dendrogram
    # -------------
    cmap = cm.rainbow(np.linspace(0, 1, 10))
    hierarchy.set_link_color_palette([mpl.colors.rgb2hex(rgb[:3]) for rgb in cmap])

    x0 = ax.get_position().x0
    x1 = ax.get_position().x1
    y0 = ax.get_position().y0
    y1 = ax.get_position().y1

    buf = y1*0.015
    dax = fig.add_axes([x0, y1+buf, x1-x0, 0.15])

    dres = shc.dendrogram(
        d_linkage,
        p=p,
        ax=dax,
        above_threshold_color="grey",
        color_threshold=color_thresh_scale*max(d_linkage[:,2])
    )

    dax.set_xticks([])
    dax.set_yticks([])
    [dax.spines[x].set_visible(False) for x in ['top','right','bottom','left']]

    # -------------
    # Metadata Axes
    # -------------

    # Agglomerative Clustering
    cluster = AgglomerativeClustering(
        n_clusters=n_clusters,
        affinity=metric,
        linkage=method
    )

    clusters = cluster.fit_predict(cmatrix.iloc[dgram_idx,dgram_idx])
    cluster_color_list, _ = series_to_colors(pd.Series(clusters))

    # Add axes
    # Plots agglomerative clustering results
    if metas is None:
        lax = fig.add_axes([x0-3*buf, y0, 2*buf, y1-y0])
        mat, cmap = color_list_to_matrix_and_cmap(cluster_color_list)
        sns.heatmap(mat.T, cmap=cmap, ax=lax, xticklabels=False, yticklabels=False, cbar=False, rasterized=True)

        uniq, idx, num_vals = np.unique(clusters.T, return_index=True, return_counts=True)
        y_locs = idx + num_vals / 2

        for idx,u in enumerate(uniq):
            lax.text(x0-50*buf, y_locs[idx], u, ha='center')

        for idx,u in enumerate(uniq):
            ax.text(
                mat.shape[1]+0.01*mat.shape[1],
                y_locs[idx],
                "n={}".format(num_vals[idx]),
                ha='left',
                fontsize=14
            )

        for _, spine in lax.spines.items():
            spine.set_visible(True)

        lax.set_xlabel("Consensus", rotation=45)

    else:
        for idx,meta in enumerate(metas):
            new_ax = [x0-(idx+3)*buf-(idx*2)*buf, y0, 2*buf, y1-y0]
            lax = fig.add_axes(new_ax)

            if isinstance(meta, str) and meta=='aggr':
                mat, cmap = color_list_to_matrix_and_cmap(cluster_color_list)
                sns.heatmap(mat.T, cmap=cmap, ax=lax, xticklabels=False, yticklabels=False, cbar=False, rasterized=True)

                uniq, idx, num_vals = np.unique(clusters.T, return_index=True, return_counts=True)
                y_locs = idx + num_vals / 2

                for idx,u in enumerate(uniq):
                    lax.text(0.5, y_locs[idx], u, ha='center')

                for idx,u in enumerate(uniq):
                    ax.text(
                        mat.shape[1]+0.01*mat.shape[1],
                        y_locs[idx],
                        "n={}".format(num_vals[idx]),
                        ha='left',
                        fontsize=14
                    )

                lax.set_xlabel("Consensus", rotation=45)


            else:
                meta = meta.loc[cmatrix.index[dgram_idx]]

                cluster_color_list, _ = series_to_colors(meta)
                mat,cmap = color_list_to_matrix_and_cmap(cluster_color_list)
                sns.heatmap(mat.T, cmap=cmap, ax=lax, yticklabels=False, xticklabels=False, cbar=False)
                lax.set_xlabel(meta.name, rotation=45)

            for _, spine in lax.spines.items():
                spine.set_visible(True)

    rs = pd.DataFrame(clusters, index=cmatrix.index[dgram_idx]).rename(columns={0:'clusters'})

    for _, spine in ax.spines.items():
        spine.set_visible(True)

    ax.set_xlabel("Samples", fontsize=14)

    return fig, rs



import pandas as pd
import numpy as np
import sys
import argparse
import time
from scipy.special import gamma
import os
import pickle
import torch
import NMF_functions
from ARD_NMF import ARD_NMF
import feather
from ARD_NMF import run_method_engine
import torch.nn as nn
import torch.multiprocessing as mp

def createFolder(directory):
    try:
        if not os.path.exists(directory):
            os.makedirs(directory)
    except OSError:
        print ('Error: Creating directory. ' +  directory)

def run_parameter_sweep(parameters, data, args, Beta):
    output = []
    num_processes = torch.cuda.device_count()
    batches = int(len(parameters) / num_processes)
    idx = 0
    objectives = []
    nsigs = []
    while idx <= len(parameters)-num_processes:
        print(idx)
        pipe_list = []
        processes = []
        for rank in range(num_processes):
            recv_end, send_end = mp.Pipe(False)
            p = mp.Process(target=run_method_engine, args=(
                data,
                parameters.iloc[idx+rank]['a'],
                parameters.iloc[idx+rank]['phi'],
                parameters.iloc[idx+rank]['b'],
                Beta,
                args.prior_on_W,
                args.prior_on_H,
                parameters.iloc[idx+rank]['K0'],
                args.tolerance,
                args.max_iter,
                args.report_frequency,
                1e-5,
                send_end,
                rank,
                ))
            pipe_list.append(recv_end)
            processes.append(p)
            p.start()

        result_list = [x.recv() for x in pipe_list]
        for p in processes:
            p.join()

        nsig = [write_output(
            x[0],
            x[1],
            data.channel_names,
            data.sample_names,
            args.output_dir,
            parameters['label'][idx+i]
            ) for i,x in enumerate(result_list)]

        [nsigs.append(ns) for i,ns in enumerate(nsig)]
        [objectives.append(x[2]) for i,x in enumerate(result_list)]
        idx += num_processes

    if idx < len(parameters):
        for i in range(len(parameters)-idx):
            idx+=i
            x = run_method_engine(
                data,
                parameters.iloc[idx]['a'],
                parameters.iloc[idx]['phi'],
                parameters.iloc[idx]['b'],
                Beta,
                args.prior_on_W,
                args.prior_on_H,
                parameters.iloc[idx]['K0'],
                args.tolerance,
                args.max_iter,
                args.report_frequency,
                1e-5,
                send_end,
                rank
            )

            nsig = write_output(
                x[0],
                x[1],
                data.channel_names,
                data.sample_names,
                args.output_dir,
                parameters['label'][idx])

            nsigs.append(nsig)
            objectives.append(x[2])

    parameters['nsigs'] = nsigs
    parameters['objective'] = objectives
    parameters.to_csv(args.output_dir + '/parameters_with_results.txt',sep='\t',index=None)

def write_output(W, H, channel_names, sample_names, output_directory, label, active_thresh = 1e-5):
    createFolder(output_directory)
    nonzero_idx = (np.sum(H, axis=1) * np.sum(W, axis=0)) > active_thresh
    W_active = W[:, nonzero_idx]
    H_active = H[nonzero_idx, :]
    nsig = np.sum(nonzero_idx)
    # Normalize W and transfer weight to H matrix
    W_weight = np.sum(W_active, axis=0)
    W_final = W_active / W_weight
    H_final = W_weight[:, np.newaxis] * H_active

    sig_names = ['W' + str(j) for j in range(1, nsig + 1)]
    W_df = pd.DataFrame(data=W_final, index=channel_names, columns=sig_names)
    H_df = pd.DataFrame(data=H_final, index=sig_names, columns=sample_names);

    # Write W and H matrices
    W_df.to_csv(output_directory + '/'+label+ '_W.txt', sep='\t')
    H_df.to_csv(output_directory + '/'+label+ '_H.txt', sep='\t')

    return nsig

def main():
    ''' Run ARD NMF'''
    torch.multiprocessing.set_start_method('spawn')

    parser = argparse.ArgumentParser(
        description='NMF with some sparsity penalty described https://arxiv.org/pdf/1111.6085.pdf')
    parser.add_argument('--data', help='Data Matrix', required=True)
    parser.add_argument('--feather', help='Input in feather format', required=False, default=False, action='store_true')
    parser.add_argument('--parquet', help='Input in parquet format', required=False, default=False, action='store_true')
    parser.add_argument('--K0', help='Initial K parameter', required=False, default=None, type=int)
    parser.add_argument('--max_iter', help='maximum iterations', required=False, default=10000, type=int)
    parser.add_argument('--del_', help='Early stop condition based on lambda change', required=False, default=1,
                        type=int)
    parser.add_argument('--tolerance', help='Early stop condition based on max lambda entry', required=False, default=1e-6,
                        type=float)
    parser.add_argument('--phi', help='dispersion parameter see paper for discussion of choosing phi '
                                      'default = 1', required=False, default=1.0, type=float)
    parser.add_argument('--a', help='Hyperparamter for lambda. We recommend trying various values of a. Smaller values'
                                    'will result in sparser results a good starting point might be'
                                    'a = log(F+N)', required=False, default=10.0,type=float)

    parser.add_argument('--b', help='Hyperparamter for lambda. Default used is as recommended in Tan and Fevotte 2012',
                        required = False,type=float, default = None)
    parser.add_argument('--objective',help='Defines the data objective. Choose between "poisson" or "gaussian". Defaults to Poisson',
                        required=False,default='poisson',type=str)

    parser.add_argument('--prior_on_W',help = 'Prior on W matrix "L1" (exponential) or "L2" (half-normal)'
                        ,required = False, default = 'L1',type=str)
    parser.add_argument('--prior_on_H',help = 'Prior on H matrix "L1" (exponential) or "L2" (half-normal)'
                        ,required = False, default = 'L1',type=str)

    parser.add_argument('--output_dir', help='output_file_name if run in array mode this correspond to the output directory', required=True)
    parser.add_argument('--labeled', help='Input has row and column labels', required=False,default=False, action='store_true')
    parser.add_argument('--report_frequency', help='Number of iterations between progress reports', required=False,
                        default=100, type=int)
    parser.add_argument('--dtype', help='Floating point accuracy', required=False,
                        default='Float32', type=str)
    parser.add_argument('--parameters_file', help='allows running many different configurations of the NMF method on a multi'
                                                  'GPU system. To run in this mode provide this argument with a text file with '
                                                  'the following headers:(a,phi,b,prior_on_W,prior_on_H,Beta,label) label '
                                                  'indicates the output stem of the results from each run.', required = False
                                                  ,default = None)
    args = parser.parse_args()


    print('Reading data frame from '+ args.data)

    if args.dtype == 'Float32':
        args.dtype = torch.float32
    elif args.dtype == 'Float16':
        args.dtype = torch.float16

    if args.parquet:
        dataset = pd.read_parquet(args.data)
    elif args.feather:
        print('loading feather...')
        dataset = feather.read_dataframe(args.data)
    else:
        if args.labeled:
            dataset = pd.read_csv(args.data, sep='\t', header=0, index_col=0)
        else:
            dataset = pd.read_csv(args.data, sep='\t', header=None)


    if args.objective.lower() == 'poisson':
        Beta = 1
    elif args.objective.lower() == 'gaussian':
        Beta = 2
    else:
        print('objective parameter should be one of "gaussian" or "poisson"')
        sys.exit()

    data = ARD_NMF(dataset, args.objective)

    if args.parameters_file != None:
        parameters = pd.read_csv(args.parameters_file, sep='\t')
        run_parameter_sweep(parameters, data, args, Beta)
    else:
        x = run_method_engine(
            data,
            args.a,
            args.phi,
            args.b,
            Beta,
            args.prior_on_W,
            args.prior_on_H,
            args.K0,
            args.tolerance,
            args.max_iter,
            args.report_frequency,
        )
        nsig = write_output(x[0],x[1],data.channel_names,data.sample_names,args.output_dir,args.output_dir)

if __name__ == "__main__":
    main()



from gprofiler import gprofiler
import pandas as pd
from typing import Union
import numpy as np
from tqdm import tqdm

import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

def gprof(
    query: Union[list, np.ndarray, pd.DataFrame],
    custom_bg: Union[list, None] = None,
    ordered_query: bool = True,
    src_filter: Union[None, str] = None,
    subtype_key: Union[None, str] = None,
    **kwargs
    ):
    """
    G-profiler wrapper for gene set enrichment analysis.
    See here: https://biit.cs.ut.ee/gprofiler/page/apis
    --------------------------------
    Gprofiler query: list of genes/proteins; can be mixed types of gene IDs, SNP IDs,
        chromosomal intervals, or term IDs

    Args:
        * query: iterable of possible Gprofiler queries or pd.DataFrame signatures output from
            ARD-NMF:

                signatures = pd.read_hdf("nmf_output.h5", "signatures")

        * custom_bg: custom background set for hypergeometric tests
            if None   --> use default background selection for gProfiler
            if (list) --> use provided genes as background
        * ordered_query (bool): g:Profiler gene lists may be interpreted as ordered
            lists where elements (genes, proteins, probesets) are in the order
            of decreasing importance. The ordered query option is useful when
            the genes can be placed in some biologically meaningful order.
            For instance, according to differential expression in a given
            RNA-seq experiment. g:Profiler then performs incremental enrichment analysis
            with increasingly larger numbers of genes starting from the top of
            the list. This procedure identifies specific functional terms that
            associate to most dramatic changes in given experimental setup, as
            well as broader terms that characterise the gene set as a whole;
            True <default>
        * src_filter (list): list of the following database sources:
            Source ID:
                GO:MF	molecular function
                GO:CC	cellular component
                GO:BP	biological process
                KEGG	Kyoto Encyclopedia of Genes and Genomes
                REAC	Reactome
                WP	    WikiPathways
                TF	    Transfac
                MIRNA	miRTarBase
                HPA	    Human Protein Atlas
                CORUM	CORUM protein complexes
                HP	    Human Phenotype Ontology
        * subtype_key (str): column to use to split input markers by; for default
            output from ARD-NMF, this will be 'max_id'; NOTE: only relevant for
            running method with pd.DataFrame

    Kwargs:
        * organism (str): 'hsapiens' <default>
        * signifcant (bool): True <default>
        * exclude_iea (bool): include electronic GO annotations: annotations assigned to
            genes using in silico curation methods adn will ahve an IEA evidence
            code to back this up; ecluding IEA annotations may help reduce bias towards
            abundant and ubiquitous housekeeping genes like ribosomal genes;
            False <default>
        * max_p_value (float): 1.0 <default>
        * max_set_size (int): 0 <default>
        * correction_method (str): 'analytical': gcustom g:SCS for reducing signifance -
            traditional multiple tests corrections are for tests independent of one another
            which is not correct for GO analysis; 'fdr': benjamini-hochberg, 'bc': bonferroni correction; 'analytical' <default>

    Returns:
        * pd.DataFrame of enrichment results
            ** Over all groupings if passed DataFrame
    """
    AVAIL_DATABASES = ("GO:MF", "GO:CC", "GO:BP", "KEGG", "REAC", "WP", "TF", "MIRNA", "HPA", "CORUM", "HP")

    if src_filter is not None:
        for db in src_filter:
            assert db in AVAIL_DATABASES, "{} is not an available database.".format(db)

    # Gprofiler syntax
    if custom_bg is None:
        custom_bg = list()

    # ----------------------------
    # Bare-bones method
    # ----------------------------
    if isinstance(query, np.ndarray) or isinstance(query, list):
        enrichment =  gprofiler(
            query,
            custom_bg=custom_bg,
            ordered_query=ordered_query,
            src_filter=src_filter,
            **kwargs
        )

        if enrichment is not None:
            return enrichment.loc[:,['p.value','term.size','overlap.size','term.name','domain','intersection']].sort_values('p.value')
        else:
            print("   * no overlaps found.")

    # ----------------------------
    # Wrapper for ARD-NMF output
    # ----------------------------
    elif isinstance(query, pd.DataFrame):
        assert subtype_key in query, "{} NOT a column in input query. Set *subtype_key* to a valid column name.".format(subtype_key)

        enrichments = list()
        for cat in tqdm(query[subtype_key].astype("category").cat.categories, desc="Computing overlaps"):
            enrichment = gprofiler(
                query[query[subtype_key]==cat].index,
                custom_bg=custom_bg,
                ordered_query=ordered_query,
                src_filter=src_filter,
                **kwargs
            )

            if enrichment is not None:
                enrichment['subtype'] = cat
                enrichment = enrichment.loc[:,['subtype','p.value','term.size','overlap.size','term.name','domain','intersection']].sort_values('p.value')
                enrichments.append(enrichment)
            else:
                print("   * no overlaps found in {}.".format(cat))

        return pd.concat(enrichments)



"""
For argument parsing and CLI interface
"""
import sys
import argparse
import os
import pkg_resources
from argparse import RawTextHelpFormatter

from .signatureanalyzer import run_maf
from .signatureanalyzer import run_spectra
from .signatureanalyzer import run_matrix

def main():
    parser = argparse.ArgumentParser(description='Signature Analyzer GPU.', formatter_class=RawTextHelpFormatter)
    parser.add_argument(
        'input',
        help="Input matrix for decomposition. Signature Analyzer uses the format of (samples x features)\n"
             "Assumes input is a .maf by default and will compute the 96-base context spectra if not provided\n"
             "  * Use {-type} to specific different input types"
    )
    parser.add_argument(
        '-t','--type',
        default='maf',
        help="Input type. Specify whether input is a .maf, a 96 base context spectra, or an RNA expression matrix (default: 'maf')\n"
             "  * NOTE: for expression is is reccomended to use log-transformed & gaussian {--objective} function",
        choices=['maf','spectra','matrix']
    )
    parser.add_argument(
        '-n','--nruns',
        help="Number of iterations to run ARD-NMF. Significant speed up if GPU is available (default: 10)",
        default=10,
        type=int
    )
    parser.add_argument(
        '-o','--outdir',
        help="Directory to save outputs (default: '.')",
        default="."
    )
    parser.add_argument(
        '--cosmic',
        help="Cosmic signatures to map to and provide results for. Support for Cosmic 2 & 3 (default: 'cosmic2')\n"
             "  * Reference: https://cancer.sanger.ac.uk/cosmic/signatures",
        default='cosmic2',
        choices=[
            'cosmic2',
            'cosmic3',
            'cosmic3_exome',
            'cosmic3_DBS',
            'cosmic3_ID',
            'cosmic3_TSB',
            ]
    )
    parser.add_argument(
        '--hg_build',
        help="Path to 2bit, human genome build for mapping mutational contexts. Required if mutational context is not provided (default: None)",
        default=None
    )
    parser.add_argument(
        '--cuda_int',
        help="GPU to use. Defaults to (cuda: 0). If (None), or if no GPU is available, will default to CPU (default: 0)",
        default=0
    )
    parser.add_argument(
        '--verbose',
        help="Verbosity",
        default=False,
        action='store_true'
    )

    # -----------------------------------------
    # NMF Options for Argparse
    # -----------------------------------------
    parser.add_argument(
        '--K0',
        help="Initial K0 parameter. If not provided, ARD-NMF starts with K0 = no. features",
        default=None,
        type=int
    )
    parser.add_argument(
        '--max_iter',
        help="Maximum number of iterations for CAVI algorithm if not reached {--tolerance} (default: 10000)",
        default=10000,
        type=int
    )
    parser.add_argument(
        '--del_',
        help="Early stop condition based on lambda change (default: 1)",
        default=1,
        type=int
    )
    parser.add_argument(
        '--tolerance',
        help="Early stop condition based on max lambda entry (default: 1e-10)",
        default=1e-10,
        type=float
    )
    parser.add_argument(
        '--phi',
        help="Dispersion parameter for CAVI. See paper for details on selection (default: 1).\n"
             "   * NOTE: If using gaussian {--objective}, scaled by the variance of input matrix",
        default=1.0,
        type=float
    )
    parser.add_argument(
        '--a',
        help="Hyperparamter for lambda. We recommend trying various values of a. Smaller values will result in\n"
             "sparser results. Reccommended starting hyperparameter: a = log(F+N). (default: 10.0)",
         default=10.0,
         type=float
     )
    parser.add_argument(
        '--b',
        help="Hyperparamter for lambda. Default is computed automatically as speicified by Tan and Fevotte 2013",
        default = None,
        type=float
    )
    parser.add_argument(
        '--objective',
        help="Objective function for ARD-NMF. (default: 'poisson')\n"
             "  * mutational signatures --> poisson (DEFAULT)\n"
             "  * log-norm expression   --> gaussian",
        default="poisson",
        choices=["poisson","gaussian"],
        type=str
    )
    parser.add_argument(
        '--prior_on_W',
        help="Prior on W matrix L1 (exponential) or L2 (half-normal) (default: 'L1')",
        default="L1",
        choices=["L1","L2"],
        type=str
    )
    parser.add_argument(
        '--prior_on_H',
        help="Prior on H matrix L1 (exponential) or L2 (half-normal) (default: 'L1')",
        default="L1",
        choices=["L1","L2"],
        type=str
    )
    parser.add_argument(
        '--report_freq',
        help="Number of iterations between progress reports (default: 250)",
        default=250,
        type=int
    )
    parser.add_argument(
        '--active_thresh',
        help="Active threshold for consdiering a threshold relevant (default: 0.01)",
        default=0.01,
        type=float
    )

    # -----------------------------------------
    # NMF Post-processing Arguments
    # -----------------------------------------
    parser.add_argument(
        '--cut_norm',
        help="Min normalized value for mean signature. Used in marker selection during post-processing (matrix). (default: 0.5)",
        default=0.5,
        type=float
    )
    parser.add_argument(
        '--cut_diff',
        help="Difference between mean selected signature and mean unselected signatures for marker selection (matrix). (default: 1.0)",
        default=1.0,
        type=float
    )

    args = parser.parse_args()
    if args.cuda_int == 'None':
        args.cuda_int = None

    print("---------------------------------------------------------")
    print("---------- S I G N A T U R E  A N A L Y Z E R  ----------")
    print("---------------------------------------------------------")

    if args.type == 'maf':
        run_maf(
            args.input,
            **vars(args)
        )
    elif args.type == 'spectra':
        run_spectra(
            args.input,
            **vars(args)
        )
    elif args.type == 'matrix':
        run_matrix(
            args.input,
            **vars(args)
        )

if __name__ == "__main__":
    main()



import itertools
import pandas as pd
from twobitreader import TwoBitFile
from typing import Union
from sys import stdout
from .utils import compl, get_true_snps_from_maf, get_dnps_from_maf

acontext = itertools.product('A', 'CGT', 'ACGT', 'ACGT')
ccontext = itertools.product('C', 'AGT', 'ACGT', 'ACGT')

context96 = dict(zip(map(''.join, itertools.chain(acontext, ccontext)), range(1, 97)))
context78 = dict(zip(['AC>CA', 'AC>CG', 'AC>CT', 'AC>GA', 'AC>GG', 'AC>GT', 'AC>TA', 'AC>TG', 'AC>TT', 'AT>CA',
                      'AT>CC', 'AT>CG', 'AT>GA', 'AT>GC', 'AT>TA', 'CC>AA', 'CC>AG', 'CC>AT', 'CC>GA', 'CC>GG',
                      'CC>GT', 'CC>TA', 'CC>TG', 'CC>TT', 'CG>AT', 'CG>GC', 'CG>GT', 'CG>TA', 'CG>TC', 'CG>TT',
                      'CT>AA', 'CT>AC', 'CT>AG', 'CT>GA', 'CT>GC', 'CT>GG', 'CT>TA', 'CT>TC', 'CT>TG', 'GC>AA',
                      'GC>AG', 'GC>AT', 'GC>CA', 'GC>CG', 'GC>TA', 'TA>AT', 'TA>CG', 'TA>CT', 'TA>GC', 'TA>GG',
                      'TA>GT', 'TC>AA', 'TC>AG', 'TC>AT', 'TC>CA', 'TC>CG', 'TC>CT', 'TC>GA', 'TC>GG', 'TC>GT',
                      'TG>AA', 'TG>AC', 'TG>AT', 'TG>CA', 'TG>CC', 'TG>CT', 'TG>GA', 'TG>GC', 'TG>GT', 'TT>AA',
                      'TT>AC', 'TT>AG', 'TT>CA', 'TT>CC', 'TT>CG', 'TT>GA', 'TT>GC', 'TT>GG'], range(1, 79)))

context83 = dict(zip(['Cdel1', 'Cdel2', 'Cdel3', 'Cdel4', 'Cdel5', 'Cdel6+',
                       'Tdel1', 'Tdel2', 'Tdel3', 'Tdel4', 'Tdel5', 'Tdel6+',
                       'Cins0', 'Cins1', 'Cins2', 'Cins3', 'Cins4', 'Cins5+',
                       'Tins0', 'Tins1', 'Tins2', 'Tins3', 'Tins4', 'Tins5+',
                       '2del1', '2del2', '2del3', '2del4', '2del5', '2del6+',
                       '3del1', '3del2', '3del3', '3del4', '3del5', '3del6+',
                       '4del1', '4del2', '4del3', '4del4', '4del5', '4del6+',
                       '5+del1', '5+del2', '5+del3', '5+del4', '5+del5', '5+del6+',
                       '2ins0', '2ins1', '2ins2', '2ins3', '2ins4', '2ins5+',
                       '3ins0', '3ins1', '3ins2', '3ins3', '3ins4', '3ins5+',
                       '4ins0', '4ins1', '4ins2', '4ins3', '4ins4', '4ins5+',
                       '5+ins0', '5+ins1', '5+ins2', '5+ins3', '5+ins4', '5+ins5+',
                       '2delm1', '3delm1', '3delm2', '4delm1', '4delm2', '4delm3',
                       '5+delm1', '5+delm2', '5+delm3', '5+delm4', '5+delm5+'], range(1, 84)))

def get_spectra_from_maf(
    maf: pd.DataFrame,
    hgfile: Union[str,None] = None,
    cosmic: str = 'cosmic2',
    real_snps: bool = False
    ):
    """
    Attaches context categories to maf and gets counts of contexts for each sample
    ---------------------------
    Args:
        * maf: Pandas DataFrame of maf
        * hgfile: path to 2bit genome build file for computing reference context
        * cosmic: cosmic signatures to decompose to

    Returns:
        * Pandas DataFrame of maf with context category attached
        * Pandas DataFrame of counts with samples as columns and context as rows
    """
    maf = maf.copy()

    if 'Start_Position' in list(maf):
        maf = maf.rename(columns={'Start_Position':'Start_position'})

    maf['sample'] = maf['Tumor_Sample_Barcode']

    if cosmic in ['cosmic2', 'cosmic3', 'cosmic3_exome']:
        # Subset to SNPs
        if 'Variant_Type' in maf.columns:
            maf = maf.loc[maf['Variant_Type'] == 'SNP']
        else:
            maf = maf.loc[maf['Reference_Allele'].apply(lambda k: len(k) == 1 and k != '-') & \
            maf['Tumor_Seq_Allele2'].apply(lambda k: len(k) == 1 and k != '-')]
        if not real_snps:
            maf = get_true_snps_from_maf(maf)

        ref = maf['Reference_Allele'].str.upper()
        alt = maf['Tumor_Seq_Allele2'].str.upper()

        if 'ref_context' in list(maf):
            context = maf['ref_context'].str.upper()
        else:
            assert hgfile is not None, 'Please provide genome build file.'

            try:
                hg = TwoBitFile(hgfile)
            except:
                raise Exception("{} not a valid 2bit file.".format(hgfile))

            # Map contexts
            _contexts = list()
            maf_size = maf.shape[0]
            for idx,(pos,chromosome) in enumerate(zip(maf["Start_position"].astype(int), maf["Chromosome"].astype(str))):
                stdout.write("\r      * Mapping contexts: {} / {}".format(idx, maf_size))

                # Double check version
                if chromosome == '23':
                    chromosome = 'X'
                elif chromosome == '24':
                    chromosome = 'Y'
                elif chromosome == 'MT':
                    chromosome = 'M'
                if not chromosome.startswith('chr'):
                    chromosome = 'chr' + chromosome

                _contexts.append(hg[chromosome][pos-2:pos+1].lower())

            maf['ref_context'] = _contexts
            stdout.write("\n")
            context = maf['ref_context'].str.upper()

        n_context = context.str.len()
        mid = n_context // 2

        contig = pd.Series([r + a + c[m - 1] + c[m + 1] if r in 'AC' \
                            else compl(r + a + c[m + 1] + c[m - 1]) \
                            for r, a, c, m in zip(ref, alt, context, mid)], index=maf.index)

        try:
            maf['context96.num'] = contig.apply(context96.__getitem__)
        except KeyError as e:
            raise KeyError('Unusual context: ' + str(e))

        maf['context96.word'] = contig
        spectra = maf.groupby(['context96.word', 'sample']).size().unstack().fillna(0).astype(int)
        for c in context96:
            if c not in spectra.index:
                spectra.loc[c] = 0
        spectra = spectra.loc[context96]

    elif cosmic == 'cosmic3_DBS':
        # Subset to DNPs
        if 'Variant_Type' not in maf.columns:
            ref_alt = maf['Reference_Allele'] + '>' + maf['Tumor_Seq_Allele2']

            def get_variant_type(ra):
                r, a = ra.split('>')
                if len(r) == 1 and r != '-' and len(a) == 1 and a != '-':
                    return 'SNP'
                if len(r) == 2 and len(a) == 2:
                    return 'DNP'
            maf['Variant_Type'] = ref_alt.apply(get_variant_type)
        if 'DNP' in maf['Variant_Type']:
            maf = maf.loc[maf['Variant_Type'] == 'DNP']
        else:
            maf = get_dnps_from_maf(maf)

        ref = maf['Reference_Allele'].str.upper()
        alt = maf['Tumor_Seq_Allele2'].str.upper()

        contig = pd.Series([r + '>' + a if r + '>' + a in context78
                            else compl(r, reverse=True) + '>' + compl(a, reverse=True)
                            for r, a in zip(ref, alt)], index=maf.index)

        try:
            maf['context78.num'] = contig.apply(context78.__getitem__)
        except KeyError as e:
            raise KeyError('Unusual context: ' + str(e))

        maf['context78.word'] = contig
        spectra = maf.groupby(['context78.word', 'sample']).size().unstack().fillna(0).astype(int)
        for c in context78:
            if c not in spectra.index:
                spectra.loc[c] = 0
        spectra = spectra.loc[context78]

    elif cosmic == 'cosmic3_ID':

        maf = maf.loc[(maf['Reference_Allele'] == '-') ^ (maf['Tumor_Seq_Allele2'] == '-')]

        ref = maf['Reference_Allele'].str.upper()
        alt = maf['Tumor_Seq_Allele2'].str.upper()

        assert hgfile is not None, 'Please provide genome build file.'

        try:
            hg = TwoBitFile(hgfile)
        except:
            raise Exception("{} not a valid 2bit file.".format(hgfile))

        # Map contexts
        contig = list()
        maf_size = maf.shape[0]
        for idx,(pos,chromosome,r,a) in enumerate(zip(maf["Start_position"].astype(int),
            maf["Chromosome"].astype(str), ref, alt)):
            stdout.write("\r      * Mapping contexts: {} / {}".format(idx, maf_size))

            # Double check version
            if chromosome == '23':
                chromosome = 'X'
            elif chromosome == '24':
                chromosome = 'Y'
            elif chromosome == 'MT':
                chromosome = 'M'
            if not chromosome.startswith('chr'):
                chromosome = 'chr' + chromosome

            if a == '-':
                del_len = len(r)
                _context = hg[chromosome][pos - 1 + del_len:pos - 1 + del_len * 6].upper()
                _context_list = [_context[n: n + del_len] for n in range(0, 5 * del_len, del_len)]
                n_repeats = 1
                for c in _context_list:
                    if c == r:
                        n_repeats += 1
                    else:
                        break
                microhomology = 0
                if n_repeats == 1:
                    for b1, b2 in zip(r, _context_list[0]):
                        if b1 == b2:
                            microhomology += 1
                        else:
                            break
                    prev_context = hg[chromosome][pos - 1 - del_len: pos - 1].upper()
                    for b1, b2 in zip(reversed(r), reversed(prev_context)):
                        if b1 == b2:
                            microhomology += 1
                        else:
                            break
                if del_len == 1:
                    pre = 'C' if r in 'CG' else 'T'
                elif del_len >= 5:
                    pre = '5+'
                else:
                    pre = str(del_len)
                if microhomology >= 5:
                    post = 'm5+'
                elif microhomology:
                    post = 'm' + str(microhomology)
                elif n_repeats == 6:
                    post = '6+'
                else:
                    post = str(n_repeats)
                contig.append(pre + 'del' + post)

            elif r == '-':
                ins_len = len(a)
                _context = hg[chromosome][pos:pos + ins_len * 5].upper()
                _context_list = [_context[n: n + ins_len] for n in range(0, 5 * ins_len, ins_len)]
                n_repeats = 0
                for c in _context_list:
                    if c == a:
                        n_repeats += 1
                    else:
                        break
                if ins_len == 1:
                    pre = 'C' if a in 'CG' else 'T'
                elif ins_len >= 5:
                    pre = '5+'
                else:
                    pre = str(ins_len)
                if n_repeats == 5:
                    post = '5+'
                else:
                    post = str(n_repeats)
                contig.append(pre + 'ins' + post)

        maf['context83.word'] = contig
        try:
            maf['context83.num'] = maf['context83.word'].apply(context83.__getitem__)
        except KeyError as e:
            raise KeyError('Unusual context: ' + str(e))

        spectra = maf.groupby(['context83.word', 'sample']).size().unstack().fillna(0).astype(int)
        for c in context83:
            if c not in spectra.index:
                spectra.loc[c] = 0
        spectra = spectra.loc[context83]

        stdout.write("\n")
    else:
        raise NotImplementedError()

    return maf, spectra



import numpy as np
import pandas as pd
import sys
import os
import matplotlib.pyplot as plt
from .utils import get_nruns_from_output

def consensus_cluster(filepath: str):
    """
    Consensus clustering of ard-nmf results.
    -----------------------
    Args:
        * filepath: path to the output .5 file of ARD-NMF runs

    Returns:
        * pd.DataFrame: consensus matrix from results
        * pd.Series: assignment probability for selected cluster

    """
    niter = get_nruns_from_output(filepath)
    H_selected = pd.read_hdf(filepath, "H")

    x = np.vstack([pd.read_hdf(filepath, "run{}/H".format(i)).loc[:,'max_id'].values for i in range(niter)])
    consensus_matrix = np.vstack([(x[:,[y]] == x[:]).sum(0) for y in range(x.shape[1])])

    df = pd.DataFrame(consensus_matrix, index=H_selected.index, columns=H_selected.index)
    df = df.loc[H_selected.sort_values('max_id').index, H_selected.sort_values('max_id').index]

    assign_p = pd.concat([df.loc[
        H_selected[H_selected['max_id']==x].index,
        H_selected[H_selected['max_id']==x].index
    ].mean(1) for x in set(H_selected['max_id'])])

    assign_p.name = 'assignment'
    return df, assign_p



"""
Signature Analyzer plotting API.
"""
from ._utils import series_to_colors
from ._utils import color_list_to_matrix_and_cmap

from ._rna import marker_heatmap

from ._muts import signature_barplot
from ._muts import signature_barplot_DBS
from ._muts import stacked_bar

from ._nmf import k_dist
from ._nmf import consensus_matrix

from ._cosine import cosine_similarity_plot



import pandas as pd
import os
from typing import Union
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import matplotlib as mpl

def series_to_colors(s: pd.Series, color_palette_map: str = 'husl', cdict: dict = None):
    """
    Convert a pandas series to a color_map.
    -----------------------
    Args:
        * s: pd.Series
        * color_palette_map: str for color map
        * cdict: dict of custom value - color mappings
            Ex. cdict = {'Low': 'Green', 'Intermediate':'Yellow', 'High': 'Red'}

    Returns:
        * pd.Series: series mapped to RGB values
        * color_map: dictionary mapping unique values in series to color
    """
    if cdict is None:
        color_labels = s.unique()
        rgb_values = sns.color_palette(color_palette_map, len(color_labels))
        color_map = dict(zip(color_labels, rgb_values))
    else:
        color_map = cdict

    for key in color_map:
        if isinstance(key, type(None)) or (isinstance(key, float) and np.isnan(key)):
            color_map[key] = (1.0,1.0,1.0,1.0)

    return s.map(color_map), color_map

def color_list_to_matrix_and_cmap(colors: list, order_dict: dict = None):
    """
    Stripped from Seaborn.
    -----------------------
    Turns a list of colors into a numpy matrix and matplotlib colormap
    These arguments can now be plotted using heatmap(matrix, cmap)
    and the provided colors will be plotted.

    Args:
        * colors : list of matplotlib colors
            Colors to label the rows or columns of a dataframe.
        * order_dict: explicit color ordering to provide
            Ex. order_dict = {'Green': 0, 'Yellow': 1, 'Red': 2}

    Returns:
        * matrix : numpy.array
            A numpy array of integer values, where each corresponds to a color
            from the originally provided list of colors
        * cmap : matplotlib.colors.ListedColormap
    """
    all_colors = set(colors)
    m = len(colors)
    colors = [colors]

    if order_dict is None:
        color_to_value = dict((col, i) for i, col in enumerate(all_colors))
    else:
        color_to_value = order_dict

    matrix = np.array([color_to_value[c] for color in colors for c in color])
    matrix = matrix.reshape((1,m))
    return matrix, mpl.colors.ListedColormap(color_to_value.keys())



import itertools
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
from typing import Union
import numpy as np
import re

from ..utils import compl, sbs_annotation_converter
from ..spectra import context96, context78, context83

def stacked_bar(H: pd.DataFrame, figsize: tuple = (8,8)):
    """
    Plot stacked barchart & normalized stacked barchart.
    --------------------------------------
    Args:
        * H: matrix output from NMF
        * figsize: size of figure (int,int)

    Returns:
        * figure

    Example usage:
        plot_bar(H)
    """
    H = H.iloc[:,:-3].copy()
    H['sum'] = H.sum(1)
    H = H.sort_values('sum', ascending=False)

    fig,axes = plt.subplots(2,1,figsize=figsize, sharex=True)

    H.iloc[:,:-1].plot(
        kind='bar',
        stacked=True,
        ax=axes[0],
        width=1.0,
        rasterized=True
    )

    axes[0].set_xticklabels([])
    axes[0].set_xticks([])
    axes[0].set_ylabel('Counts', fontsize=20)

    H_norm = H.iloc[:,:-1].div(H['sum'].values,axis=0)
    H_norm.plot(
        kind='bar',
        stacked=True,
        ax=axes[1],
        width=1.0,
        rasterized=True
    )

    axes[1].set_xticklabels([])
    axes[1].set_xticks([])
    axes[1].set_xlabel('Samples', fontsize=16)
    axes[1].set_ylabel('Fractions', fontsize=20)
    axes[1].get_legend().remove()
    axes[1].set_ylim([0,1])

    return fig

def _map_sbs_sigs_back(df: pd.DataFrame) -> pd.Series:
    """
    Map Back Single-Base Substitution Signatures.
    -----------------------
    Args:
        * df: pandas.core.frame.DataFrame with index to be mapped

    Returns:
        * pandas.core.series.Series with matching indices to context96
    """
    def _check_to_flip(x, ref):
        if x in ref:
            return x
        else:
            return compl(x)

    if df.index.name is None: df.index.name = 'index'
    df_idx = df.index.name

    if ">" in df.index[0]:
        # Already in word format
        context_s = df.reset_index()[df_idx].apply(sbs_annotation_converter)
    else:
        # Already in arrow format
        context_s = df.reset_index()[df_idx]

    return context_s.apply(lambda x: _check_to_flip(x, context96.keys()))

def signature_barplot(W: pd.DataFrame, contributions: Union[int, pd.Series] = 1):
    """
    Plots signatures from W-matrix for Single-Base Substitutions
    --------------------------------------
    Args:
        * W: W-matrix
        * contributions: Series of total contributions, np.sum(H), from each
            signature if W is normalized; else, 1

    Returns:
        * fig

    Example usage:
        signature_barplot(W, np.sum(H))
    """
    W = W.copy()
    W.index = _map_sbs_sigs_back(W)

    for c in context96:
        if c not in W.index:
            W.loc[c] = 0

    W.sort_index(inplace=True)

    sig_columns = [c for c in W if c.startswith('S')]

    if isinstance(contributions, pd.Series):
        W = W[sig_columns] * contributions[sig_columns]
    else:
        W = W[sig_columns] * contributions

    n_sigs = len(sig_columns)

    context_label = []
    change_map = {'CA': [], 'CG': [], 'CT': [], 'TA': [], 'TC': [], 'TG': []}
    for p in itertools.product('ACGT', 'ACGT'):
        context = ''.join(p)
        compl_context = compl(context, reverse=True)
        context_label.append('-'.join(context))
        for key in change_map:
            if key.startswith('C'):
                change_map[key].append(key + context)
            else:
                change_map[key].append(compl(key) + compl_context)
    color_map = {'CA': 'cyan', 'CG': 'red', 'CT': 'yellow', 'TA': 'purple', 'TC': 'green', 'TG': 'blue'}

    x_coords = range(16)
    fig, axes = plt.subplots(nrows=n_sigs, ncols=6, figsize=(20, 2.5 * n_sigs), sharex='col', sharey='row')
    for row, sig in enumerate(sig_columns):
        for col, chg in enumerate(['CA', 'CG', 'CT', 'TA', 'TC', 'TG']):
            if n_sigs == 1:
                ax = axes[col]
            else:
                ax = axes[row, col]
            bar_heights = W[sig].loc[change_map[chg]]
            ax.bar(x_coords, bar_heights, width=.95, linewidth=1.5, edgecolor='gray', color=color_map[chg], rasterized=True)
            ax.set_xlim(-.55, 15.55)
            if row == 0:
                ax.set_title('>'.join(chg), fontsize=18)
                if col == 0:
                    ax.text(51.2 / 16, 1.3, 'Mutational Signatures', transform=ax.transAxes,
                            horizontalalignment='center', fontsize=24)
            if row < n_sigs - 1:
                ax.tick_params(axis='x', length=0)
            else:
                ax.set_xticks(x_coords)
                ax.set_xticklabels(context_label, fontfamily='monospace', rotation='vertical')
                if col == 0:
                    ax.text(51.2 / 16, -.4, 'Motifs', transform=ax.transAxes, horizontalalignment='center', fontsize=20,
                            fontweight='bold')
            if col > 0:
                ax.tick_params(axis='y', length=0)
            if col == 5:
                ax.text(1.05, .5, sig, fontsize=14, rotation=270, transform=ax.transAxes, verticalalignment='center')

    plt.subplots_adjust(wspace=.08, hspace=.15)
    fig.text(.08, .5, 'Contributions', rotation='vertical', verticalalignment='center', fontsize=20, fontweight='bold')

    return fig

def signature_barplot_DBS(W, contributions):
    """
    Plots signatures from W-matrix for Doublet-Base Substitutions
    --------------------------------------
    Args:
        * W: W-matrix
        * contributions: Series of total contributions, np.sum(H), from each
            signature if W is normalized; else, 1

    Returns:
        * fig

    Example usage:
        signature_barplot_DBS(W, np.sum(H))
    """
    W = W.copy()
    for c in context78:
        if c not in W.index:
            W.loc[c] = 0
    W.sort_index(inplace=True)
    sig_columns = [c for c in W if c.startswith('S')]
    if isinstance(contributions, pd.Series):
        W = W[sig_columns] * contributions[sig_columns]
    else:
        W = W[sig_columns] * contributions

    n_sigs = len(sig_columns)

    ref_map = {'AC': [], 'AT': [], 'CC': [], 'CG': [], 'CT': [], 'GC': [], 'TA': [], 'TC': [], 'TG': [], 'TT': []}
    for x in W.index:
        ref_map[x[:2]].append(x)
    x_coords = {ref: range(len(sigs)) for ref, sigs in ref_map.items()}

    color_map = {'AC': '#99CCFF', 'AT': '#0000FF', 'CC': '#CCFF99', 'CG': '#00FF00', 'CT': '#FF99CC',
                 'GC': '#FF0000', 'TA': '#FFCC99', 'TC': '#FF8000', 'TG': '#CC99FF', 'TT': '#8000FF'}
    fig, axes = plt.subplots(nrows=n_sigs, ncols=10, figsize=(20, 2.5 * n_sigs), sharex='col',
                             sharey='row', gridspec_kw={'width_ratios': (3, 2, 3, 2, 3, 2, 2, 3, 3, 3)})
    for row, sig in enumerate(sig_columns):
        for col, ref in enumerate(ref_map):
            if n_sigs == 1:
                ax = axes[col]
            else:
                ax = axes[row, col]
            bar_heights = W[sig].loc[ref_map[ref]]
            ax.bar(x_coords[ref], bar_heights, width=.95, linewidth=1.5, edgecolor='gray', color=color_map[ref],
                   rasterized=True)
            ax.set_xlim(-.55, x_coords[ref][-1] + .55)
            if row == 0:
                ax.set_title(ref)
                if col == 0:
                    ax.text(44.5 / 6, 1.2, 'Mutational Signatures', transform=ax.transAxes,
                            horizontalalignment='center', fontsize=24)
            if row < n_sigs - 1:
                ax.tick_params(axis='x', length=0)
            else:
                xlabels = [x[3:] for x in ref_map[ref]]
                ax.set_xticks(x_coords[ref])
                ax.set_xticklabels(xlabels, fontfamily='monospace', rotation='vertical')
                if col == 0:
                    ax.text(44.5 / 6, -.3, 'Motifs', transform=ax.transAxes, horizontalalignment='center', fontsize=20,
                            fontweight='bold')
            if col > 0:
                ax.tick_params(axis='y', length=0)
            if col == 9:
                ax.text(1.05, .5, sig, fontsize=14, rotation=270, transform=ax.transAxes, verticalalignment='center')

    plt.subplots_adjust(wspace=.08, hspace=.15)
    fig.text(.08, .5, 'Contributions', rotation='vertical', verticalalignment='center', fontsize=20, fontweight='bold')

    return fig

def plot_signatures_ID(W, contributions):
    """
    Plots signatures from W-matrix for Insertions-Deletions
    --------------------------------------
    Args:
        * W: W-matrix
        * contributions: Series of total contributions, np.sum(H), from each
            signature if W is normalized; else, 1

    Returns:
        * fig

    Example usage:
        signature_barplot_ID(W, np.sum(H))
    """
    W = W.copy()
    for c in context83:
        if c not in W.index:
            W.loc[c] = 0
    W = W.loc[context83]
    sig_columns = [c for c in W if c.startswith('S')]
    if isinstance(contributions, pd.Series):
        W = W[sig_columns] * contributions[sig_columns]
    else:
        W = W[sig_columns] * contributions

    n_sigs = len(sig_columns)
    group_map = {'Cdel': [], 'Tdel': [], 'Cins': [], 'Tins': [],
                 '2del': [], '3del': [], '4del': [], '5+del': [],
                 '2ins': [], '3ins': [], '4ins': [], '5+ins': [],
                 '2delm': [], '3delm': [], '4delm': [], '5+delm': []}
    for x in W.index:
        group = re.search('.+?(?=[\d])', x).group(0)
        group_map[group].append(x)
    x_coords = {group: range(len(sigs)) for group, sigs in group_map.items()}

    color_map = {'Cdel': '#FFCC99', 'Tdel': '#FF8000', 'Cins': '#00FF00', 'Tins': '#00BB00',
                 '2del': '#FF99CC', '3del': '#FF3377', '4del': '#FF0000', '5+del': '#880000',
                 '2ins': '#99CCFF', '3ins': '#3377FF', '4ins': '#0000FF', '5+ins': '#000088',
                 '2delm': '#CC99FF', '3delm': '#9966FF', '4delm': '#8000FF', '5+delm': '#6000AA'}

    fig, axes = plt.subplots(nrows=n_sigs, ncols=16, figsize=(20, 2.5 * n_sigs), sharex='col',
                             sharey='row', gridspec_kw={'width_ratios': (6,) * 12 + (1, 2, 3, 5)})
    for row, sig in enumerate(sig_columns):
        for col, group in enumerate(group_map):
            if n_sigs == 1:
                ax = axes[col]
            else:
                ax = axes[row, col]
            bar_heights = W[sig].loc[group_map[group]]
            ax.bar(x_coords[group], bar_heights, width=.95, linewidth=1.5, edgecolor='gray', color=color_map[group],
                   rasterized=True)
            ax.set_xlim(-.55, x_coords[group][-1] + .55)
            if row == 0:
                ax.set_title(re.search('[\d+CT]+', group).group(0), color=color_map[group])
                if col == 0:
                    ax.text(44.5 / 6, 1.3, 'Mutational Signatures', transform=ax.transAxes,
                            horizontalalignment='center', fontsize=24)
                if group == 'Tdel':
                    ax.text(-.02, 1.16, '1bp deletions at repeats', fontsize=10, transform=ax.transAxes,
                            horizontalalignment='center', color=color_map[group])
                if group == 'Tins':
                    ax.text(-.02, 1.16, '1bp insertions at repeats', fontsize=10, transform=ax.transAxes,
                            horizontalalignment='center', color=color_map[group])
                if group == '4del':
                    ax.text(-.02, 1.16, '>1bp deletions at repeats', fontsize=10, transform=ax.transAxes,
                            horizontalalignment='center', color=color_map[group])
                if group == '4ins':
                    ax.text(-.02, 1.16, '>1bp insertions at repeats', fontsize=10, transform=ax.transAxes,
                            horizontalalignment='center', color=color_map[group])
                if group == '4delm':
                    ax.text(.8, 1.16, '>1bp deletions with microhomology', fontsize=10, transform=ax.transAxes,
                            horizontalalignment='center', color=color_map[group])
            if row < n_sigs - 1:
                ax.tick_params(axis='x', length=0)
            else:
                xlabels = [re.search('[\d+]+$', x).group(0) for x in group_map[group]]
                ax.set_xticks(x_coords[group])
                ax.set_xticklabels(xlabels, fontfamily='monospace')
                if col == 0:
                    ax.text(44.5 / 6, -.3, 'Motifs', transform=ax.transAxes, horizontalalignment='center', fontsize=20,
                            fontweight='bold')
            if col > 0:
                ax.tick_params(axis='y', length=0)
            if col == 15:
                ax.text(1.05, .5, sig, fontsize=14, rotation=270, transform=ax.transAxes, verticalalignment='center')

    plt.subplots_adjust(wspace=.08, hspace=.15)
    fig.text(.08, .5, 'Contributions', rotation='vertical', verticalalignment='center', fontsize=20, fontweight='bold')

    return fig



from setuptools import setup
import re
import os
import sys

ver_info = sys.version_info
if ver_info < (3,6,0):
    raise RuntimeError("signatureanalyzer requires at least python 3.6.0")

with open(os.path.join(os.path.dirname(__file__), 'signatureanalyzer', '__init__.py')) as r:
    version = re.search(r'__version__ = \'(\d+\.\d+\.\d+[-_a-zA-Z0-9]*)\'', r.read()).group(1)

setup(
    name = 'signatureanalyzer',
    version = version,
    author = 'Shankara Anand & Justin Cha - Broad Institute - Cancer Genome Computational Analysis',
    author_email = 'sanand@broadinstitute.org',
    url = 'https://github.com/broadinstitute/getzlab-SignatureAnalyzer',
    long_description = open("README.md", encoding="utf-8").read(),
    long_description_content_type = 'text/markdown',
    description = 'Bayesian NMF methods for mutational signature analysis & transcriptomic profiling on GPUs (Getz Lab).',
    packages = [
        'signatureanalyzer',
        'signatureanalyzer.plotting',
        'signatureanalyzer.pathways',
        'signatureanalyzer.signatureanalyzer_gpu'
    ],
    install_requires = [
        "twobitreader>=3.1.7",
        "torch>=1.2.0",
        "seaborn>=0.9.0",
        "pandas>=0.25.0",
        "pyarrow>=0.14.1",
        "scikit-learn>=0.21.3",
        "scikit-image>=0.15.0",
        "tqdm>=4.33.0",
        "h5py>=2.9.0",
        "tables>=3.6.1",
        "missingpy",
        "gprofiler",
        "numpy",
        "matplotlib",
        "scipy"
    ],
    package_data = {
    "":[
        "ref/cosmic_v2/sa*",
        "ref/cosmic_v3/sa*"
    ]
    },
    entry_points = {
        'console_scripts': [
            'signatureanalyzer = signatureanalyzer.__main__:main'
        ]
    },
    classifiers = [
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
    ],
    license="MIT"
)



# Modules
from . import plotting as pl
from . import pathways as pw

# Independent imports
from .bnmf import ardnmf
from .utils import postprocess_msigs
from .consensus import consensus_cluster

from .signatureanalyzer import run_maf
from .signatureanalyzer import run_spectra
from .signatureanalyzer import run_matrix

__version__ = '0.0.5'



import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import scipy.cluster
import pandas as pd

def cosine_similarity_plot(cosine_similarity_matrix: pd.DataFrame):
    """
    Plots cosine similarity heatmap
    --------------------------------------
    Args:
        * cosine_similarity_matrix: DataFrame with cosine similarities of each component to each signature

    Returns:
        * fig

    Example usage:
        cosine_similarity_plot(pd.read_hdf('nmf_output.h5', 'cosine'))
    """

    linkage_mat = scipy.cluster.hierarchy.linkage(cosine_similarity_matrix, metric='euclidean')
    sig_order = scipy.cluster.hierarchy.dendrogram(linkage_mat, no_plot=True)['leaves']
    sorted_cosine_similarity_matrix = cosine_similarity_matrix.iloc[sig_order]

    fig, (ax, cbar_ax) = plt.subplots(figsize=(cosine_similarity_matrix.shape[1] + .5, cosine_similarity_matrix.shape[0] / 3),
                                      ncols=2, gridspec_kw={'width_ratios': [cosine_similarity_matrix.shape[1] * 2, 1]})
    sns.heatmap(sorted_cosine_similarity_matrix, vmin=.75, vmax=1, ax=ax, cbar_ax=cbar_ax, cmap='viridis', annot=True)
    ax.set_yticks(np.arange(cosine_similarity_matrix.shape[0]) + .5)
    ax.set_yticklabels(sorted_cosine_similarity_matrix.index)
    ymax, ymin = ax.get_ylim()
    ax.set_ylim(ymax + .5, ymin - .5)
    for _, spine in ax.spines.items():
        spine.set_visible(True)
    cbar_ax.set_frame_on(True)

    return fig



import numpy as np
import pandas as pd
import sys
from tqdm import tqdm
import h5py
from sklearn.metrics.pairwise import cosine_similarity
import pkg_resources

from missingpy import KNNImputer, MissForest

COMPL = {"A":"T","T":"A","G":"C","C":"G"}

# ---------------------------------
# IOUtils
# ---------------------------------
def file_loader(x):
    if x.endswith('.csv'):
        return pd.read_csv(x, index_col=0)
    elif x.endswith('.parquet'):
        return pd.read_parquet(x)
    else:
        return pd.read_csv(x, sep='\t', index_col=0)

# ---------------------------------
# NMF Utils
# ---------------------------------
def split_negatives(x: pd.DataFrame, tag: str = '_n', axis: int = 0):
    """
    Split dataframe into positive and negative components.
    --------------------
    Args:
        * x: pd.DataFrame input matrix
        * tag: string that will be added to the end of the negative variable names
        * axis: which axis to create a positive dimension for
            NOTE: this is for 2D numpy arrays

    Returns:
        * pd.DataFrame with new positive transformed matrix.
    """
    x_neg = -1 * x.copy()
    x_neg = x_neg.where(x_neg > 0, 0)

    if axis:
        x.columns = x.columns.astype(str)
        x_neg.columns = [x+'_n' for x in x.columns]
    else:
        x.index = x.index.astype(str)
        x_neg.index = [x+'_n' for x in x.index]

    return pd.concat([x.where(x > 0, 0), x_neg], axis=axis)

def impute_values(df: pd.DataFrame, method: str = 'mean', **kwargs):
    """
    Impute missing values in DataFrame (np.nan or None).
    ------------------------
    Args:
        * df: pd.DataFrame of (samples x features)
        * method: string for what method of imputation to use
            ** 'mean': mean imputation
            ** 'knn': K-NN imputation (see missingpy.KNNImputer)
            ** 'rf': random forest imputation (see missingpy.MissForest)

    Returns:
        * pd.DataFrame: imputed values (samples x features)
    """
    assert method in ('mean','knn','rf'), '{} not yet implemented.'.format(method)

    if method=='mean':
        return df.fillna(df.mean(0))
    elif method=='knn':
        X = df.values
        imputer = KNNImputer(**kwargs)
        X_impute = imputer.fit_transform(X)
        return pd.DataFrame(X_impute, index=df.index, columns=df.columns)
    elif method=='rf':
        X = df.values
        imputer = MissForest(**kwargs)
        X_impute = imputer.fit_transform(X)
        return pd.DataFrame(X_impute, index=df.index, columns=df.columns)

def l2fc(df: pd.DataFrame, center: str = 'median', axis: int = 1):
    """
    Log2 Fold-Change Input Dataframe
    -------------------------
    Args:
        * df: pd.DataFrame
        * center: center-metrix to compute log-fold change over
            ** 'median'
            ** 'mean'
        * axis: int axis to compute median and LFC across (i.e. samples)

    Returns:
        * pd.Dataframe: log2FC-transformed pandas dataframe
    """
    X = df.values

    if center == 'median':
        X_mid = np.median(X, axis)
    elif center == 'mean':
        X_mid = np.mean(X, axis)

    if axis==1:
        return pd.DataFrame(np.log2(X) - np.log2(X_mid)[:,np.newaxis], index=df.index, columns=df.columns)
    else:
        return pd.DataFrame(np.log2(X) - np.log2(X_mid)[np.newaxis], index=df.index, columns=df.columns)

def compute_phi(mu: float, var: float, beta: float):
    """
    Compute Phi
    ------------------------
    Compute the Dispersion parameter.
    """
    return var / (mu ** (2-beta))

def transfer_weights(W: pd.DataFrame, H: pd.DataFrame, active_thresh:float = 1e-2):
    """
    Transfers weights from output of NMF.
    ------------------------
    Args:
        * W: input W matrix (K x n_features)
        * H: input H matrix (n_samples x K)
        * active_thresh: active threshold to consider a factor loading significant

    Returns:
        * W_final: normalized W matrix (K x n_features)
        * H_final: normalized H matrix (n_samples x K)
        * nsig: number of signatures found
    """
    W = W.copy()
    H = H.copy()

    nonzero_idx = (np.sum(H, axis=1) * np.sum(W, axis=0)) > active_thresh

    W_active = W[:, nonzero_idx]
    H_active = H[nonzero_idx, :]
    nsig = np.sum(nonzero_idx)

    # Normalize W and transfer weight to H matrix
    W_weight = np.sum(W_active, axis=0)
    W_final = W_active / W_weight
    H_final = W_weight[:, np.newaxis] * H_active

    return W_final, H_final, nsig, nonzero_idx

def select_signatures(W: pd.DataFrame, H: pd.DataFrame):
    """
    Scales NMF output by sample and feature totals to select Signatures.
    ------------------------
    Args:
        * W: input W matrix (K x n_features)
        * H: input H matrix (n_samples x K)

    Returns:
        * W: output W matrix with max_id, max, and max_norm columns
        * H: output H matrix with max_id, max, and max_norm columns
    """
    Wnorm = W.copy()
    Hnorm = H.copy()

    # Scale Matrix
    for j in range(W.shape[1]):
        Wnorm.iloc[:,j] *= H.sum(1).values[j]
        Hnorm.iloc[j,:] *= W.sum(0).values[j]

    # Normalize
    Wnorm = Wnorm.div(Wnorm.sum(1),axis=0)
    Hnorm = Hnorm.div(Hnorm.sum(0),axis=1)

    H = H.T
    Hnorm = Hnorm.T

    # Get Max Values
    H_max_id = H.idxmax(axis=1, skipna=True).astype('int')
    H['max'] = H.max(axis=1, skipna=True)
    H['max_id'] = H_max_id
    Hnorm['max_norm']=Hnorm.max(axis=1, skipna=True)

    W_max_id = W.idxmax(axis=1, skipna=True).astype('int')
    W['max'] = W.max(axis=1, skipna=True)
    W['max_id'] = W_max_id
    Wnorm['max_norm']=Wnorm.max(axis=1, skipna=True)

    H['max_norm'] = Hnorm['max_norm']
    W['max_norm'] = Wnorm['max_norm']

    _rename = {x:'S'+x for x in list(H)[:-3]}
    H = H.rename(columns=_rename)
    W = W.rename(columns=_rename)

    return W,H

def select_markers(
    X: pd.DataFrame, \
    W: pd.DataFrame, \
    H: pd.DataFrame, \
    cut_norm: float = 0.5, \
    cut_diff: float = 1.0, \
    verbose: bool = False \
    ):
    """
    Marker selection from NMF.
    ------------------------
    Args:
        * X: Input X matrix (n_samples x n_features)
        * W: input W matrix (K x n_features) with max_id, max, and max_norm columns
        * H: input H matrix (n_samples x K) with max_id, max, and max_norm columns
        * cut_norm: minimum normalized signature strength
        * cut_diff: minimum difference between selected signature and other signatures

    Returns:
        * Pandas Dataframe of NMF markers
        * Pandas Dataframe of full W matrix
    """
    markers = list()
    full = list()

    pd.options.mode.chained_assignment = None

    for n in tqdm(np.unique(W['max_id']), desc='Clusters: ', disable=not verbose):
        if H[H['max_id']==n].shape[0] > 0:
            tmp = W[W['max_id']==n]
            tmp.loc[:,'mean_on'] = X.loc[np.array(tmp.index), H[H['max_id']==n].index].mean(axis=1)
            tmp.loc[:,'mean_off'] = X.loc[np.array(tmp.index), H[H['max_id']!=n].index].mean(axis=1)
            tmp.loc[:,'diff'] = tmp.loc[:,'mean_on'] - tmp.loc[:,'mean_off']

            tmp.sort_values('diff', ascending=False, inplace=True)
            full.append(tmp)
            markers.append(tmp[(tmp['diff'] > cut_diff) & (tmp['max_norm'] >= cut_norm)])

    nmf_markers = X.loc[pd.concat(markers).index,H.max_id.sort_values().index]
    nmf_markers.index.name = 'feat'

    return nmf_markers, pd.concat(full)

# ---------------------------------
# Mutational Signature Utils
# ---------------------------------
def load_cosmic_signatures(cosmic: str):
    """
    Load cosmic signatures.
    -------------------------
    Pre-processed Cosmic Mutational Signatures.
    """
    if cosmic == 'cosmic2':
        print("   * Using {} signatures".format(cosmic))
        cosmic = pd.read_csv(pkg_resources.resource_filename('signatureanalyzer', 'ref/cosmic_v2/sa_cosmic2.tsv'), sep='\t').dropna(1)
        cosmic_index = "Somatic Mutation Type"
    elif cosmic == 'cosmic3':
        print("   * Using {} signatures".format(cosmic))
        cosmic = pd.read_csv(pkg_resources.resource_filename('signatureanalyzer', 'ref/cosmic_v3/sa_cosmic3_sbs.tsv'), sep='\t').dropna(1)
        cosmic_index = "Somatic Mutation Type"
    elif cosmic == 'cosmic3_exome':
        print("   * Using {} signatures".format(cosmic))
        cosmic = pd.read_csv(pkg_resources.resource_filename('signatureanalyzer', 'ref/cosmic_v3/sa_cosmic3_sbs_exome.tsv'), sep='\t').dropna(1)
        cosmic_index = "Somatic Mutation Type"
    elif cosmic == 'cosmic3_DBS':
        print("   * Using {} signatures".format(cosmic))
        cosmic = pd.read_csv(pkg_resources.resource_filename('signatureanalyzer', 'ref/cosmic_v3/sa_cosmic3_dbs.tsv'), sep='\t').dropna(1)
        cosmic_index = "Somatic Mutation Type"
    else:
        raise Exception("Not yet implemented for {}".format(cosmic))

    return cosmic, cosmic_index

def compl(seq: str, reverse: bool = False):
    """
    Gets the complement of a string
    Args:
        * seq: string (does not have to be base pair)
        * reverse: set to true to reverse seq

    Returns:
        * complement of seq
    """
    return ''.join([COMPL[x] if x in COMPL.keys() else x for x in (reversed(seq) if reverse else seq)])

def sbs_annotation_converter(x: str) -> str:
    """
    Eithers swaps from word -> arrow format for SBS or vice versa.
        word: (REF)(ALT)(LEFT)(RIGHT)
        arrow: (LEFT)[(REF)>(ALT)](RIGHT)
    """
    if '>' in x:
        return x[2]+x[4]+x[0]+x[6]
    else:
        return x[2]+'['+x[0]+'>'+x[1]+']'+x[3]

def _map_id_sigs(
    df: pd.DataFrame,
    cosmic_df: pd.DataFrame,
    sub_index: str = 'Substitution Type'
    ) -> pd.Series:
    """
    Map Insertion-Deletion Substitution Signatures.
    -----------------------
    Args:
        * df: pandas.core.frame.DataFrame with index to be mapped
        * cosmic_df: dataframe with Cosmic indices to map to
        * sub_index: substitution index - the column to map to in the cosmic dataframe

    Returns:
        * pandas.core.series.Series with matching indices to input cosmic
    """
    # TODO @ Justin
    pass

def _map_dbs_sigs(
    df: pd.DataFrame,
    cosmic_df: pd.DataFrame,
    sub_index: str = 'Substitution Type'
    ) -> pd.Series:
    """
    Map Doublet-Base Substitution Signatures.
    -----------------------
    Args:
        * df: pandas.core.frame.DataFrame with index to be mapped
        * cosmic_df: dataframe with Cosmic indices to map to
        * sub_index: substitution index - the column to map to in the cosmic dataframe

    Returns:
        * pandas.core.series.Series with matching indices to input cosmic
    """
    def _check_to_flip(x, ref):
        if x in ref:
            return x
        else:
            return compl(x)

    if df.index.name is None: df.index.name = 'index'
    df_idx = df.index.name

    context_s = df.reset_index()[df_idx]
    return context_s.apply(lambda x: _check_to_flip(x, set(cosmic_df[sub_index])))

def _map_sbs_sigs(
    df: pd.DataFrame,
    cosmic_df: pd.DataFrame,
    sub_index: str = 'Substitution Type'
    ) -> pd.Series:
    """
    Map Single-Base Substitution Signatures.
    -----------------------
    Args:
        * df: pandas.core.frame.DataFrame with index to be mapped
        * cosmic_df: dataframe with Cosmic indices to map to
        * sub_index: substitution index - the column to map to in the cosmic dataframe

    Returns:
        * pandas.core.series.Series with matching indices to input cosmic
    """
    def _check_to_flip(x, ref):
        if x[2:-2] in ref:
            return x
        else:
            return compl(x)

    if df.index.name is None: df.index.name = 'index'
    df_idx = df.index.name

    if ">" not in df.index[0]:
        # Already in word format
        context_s = df.reset_index()[df_idx].apply(sbs_annotation_converter)
    else:
        # Already in arrow format
        context_s = df.reset_index()[df_idx]

    return context_s.apply(lambda x: _check_to_flip(x, set(cosmic_df[sub_index])))

def postprocess_msigs(res: dict, cosmic: pd.DataFrame, cosmic_index: str, cosmic_type: str):
    """
    Post process ARD-NMF on mutational signatures.
    ------------------------
    Args:
        * res: results dictionary from ARD-NMF (see ardnmf function)
        * cosmic: cosmic pd.DataFrmae
        * cosmic_index: feature index column in cosmic
            ** ex. in cosmic_v2, "Somatic Mutation Type" columns map to
                A[C>A]A, A[C>A]C, etc.

    Returns:
        * None, edits res dictionary directly
    """
    if cosmic_type in ('cosmic2','cosmic3','cosmic3_exome'):
        res["Wraw"]["mut"] = _map_sbs_sigs(res["Wraw"], cosmic).values
    elif cosmic_type == 'cosmic3_DBS':
        res["Wraw"].to_parquet("W_test.parquet")
        cosmic.to_parquet("cosmic_test.parquet")
        res["Wraw"]["mut"] = _map_dbs_sigs(res["Wraw"], cosmic).values

    # Column names of NMF signatures & COSMIC References
    nmf_cols = ["S"+x for x in list(map(str, set(res["signatures"].max_id)))]
    ref_cols = list(cosmic.columns[cosmic.dtypes == 'float64'])

    # Create cosine similarity matrix
    X = res["Wraw"].set_index("mut").join(cosmic.set_index(cosmic_index)).dropna(1).loc[:,nmf_cols+ref_cols]
    res["cosine"] = pd.DataFrame(cosine_similarity(X.T), index=X.columns, columns=X.columns).loc[ref_cols,nmf_cols]

    # Add assignments
    s_assign = dict(res["cosine"].idxmax())
    s_assign = {key:key+"-" + s_assign[key] for key in s_assign}

    # Update column names
    res["W"] = res["W"].rename(columns=s_assign)
    res["H"] = res["H"].rename(columns=s_assign)
    res["Wraw"] = res["Wraw"].rename(columns=s_assign)
    res["Hraw"] = res["Hraw"].T.rename(columns=s_assign)
    res["cosine"] = res["cosine"].rename(columns=s_assign)

# ---------------------------------
# Parsing Output H5 Utils
# ---------------------------------
def get_nruns_from_output(file: str) -> int:
    """
    Get number of NMF runs from an .h5 output file.
    ------------------------
    Args:
        * output file from NMF (.h5)

    Returns:
        * number of NMF runs (int)
    """
    with h5py.File(file) as f:
        return(max([int(x.split('run')[1]) for x in list(f.keys()) if 'run' in x])+1)

def get_nlogs_from_output(file: str) -> pd.DataFrame:
    """
    Returns a dataframe of the final output log from each run.
    ------------------------
    Args:
        * output file from NMF (.h5)

    Returns:
        * pd.DataFrame of reporting statistics for each run
    """
    n_runs = get_nruns_from_output(file)
    df = pd.concat([pd.read_hdf(file,"run{}/log".format(i)).reset_index().iloc[-1] for i in range(n_runs)],1).T
    df.index = np.arange(n_runs)
    return df

def get_dnps_from_maf(maf: pd.DataFrame):
    sub_mafs = []
    for _, df in maf.loc[maf['Variant_Type'] == 'SNP'].groupby(['sample', 'Chromosome']):
        df = df.sort_values('Start_position')
        start_pos = np.array(df['Start_position'])
        pos_diff = np.diff(start_pos)
        idx = []
        if len(pos_diff) == 1 and pos_diff[0] == 1:
            idx.append(0)
        if len(pos_diff) >= 2 and pos_diff[0] == 1 and pos_diff[1] > 1:
            idx.append(0)
        idx.extend(np.flatnonzero((pos_diff[:-2] > 1) & (pos_diff[1:-1] == 1) & (pos_diff[2:] > 1)) + 1)
        if len(pos_diff) >= 2 and pos_diff[-1] == 1 and pos_diff[-2] > 1:
            idx.append(len(pos_diff) - 1)
        if idx:
            idx = np.array(idx)
            rows = df.iloc[idx][['Hugo_Symbol', 'Tumor_Sample_Barcode', 'sample',
                'Chromosome', 'Start_position', 'Reference_Allele', 'Tumor_Seq_Allele2']].reset_index(drop=True)
            rows_plus_one = df.iloc[idx + 1].reset_index()
            rows['Variant_Type'] = 'DNP'
            rows['End_position'] = rows['Start_position'] + 1
            rows['Reference_Allele'] = rows['Reference_Allele'] + rows_plus_one['Reference_Allele']
            rows['Tumor_Seq_Allele2'] = rows['Tumor_Seq_Allele2'] + rows_plus_one['Tumor_Seq_Allele2']
            sub_mafs.append(rows)
    return pd.concat(sub_mafs).reset_index(drop=True)

def get_true_snps_from_maf(maf: pd.DataFrame):
    sub_mafs = []
    for _, df in maf.loc[maf['Variant_Type'] == 'SNP'].groupby(['sample', 'Chromosome']):
        df = df.sort_values('Start_position')
        start_pos = np.array(df['Start_position'])
        pos_diff = np.diff(start_pos)
        rem_idx = np.flatnonzero(pos_diff <= 1)
        rem_idx = np.concatenate([rem_idx, rem_idx + 1])
        idx = np.delete(np.arange(len(start_pos)), rem_idx)
        if len(idx):
            rows = df.iloc[idx]
            sub_mafs.append(rows)
    return pd.concat(sub_mafs).reset_index(drop=True)



import numpy as np
import pandas as pd
import sys
import os
import torch
from typing import Union

# sys.path.append(os.path.join(os.path.dirname(__file__), '.', 'SignatureAnalyzer-GPU'))
# from ARD_NMF import ARD_NMF, run_method_engine

from .signatureanalyzer_gpu.ARD_NMF import ARD_NMF, run_method_engine

# Relative Imports
from .utils import compute_phi, transfer_weights, select_signatures, select_markers


# ---------------------------------
# ARD - NMF Wrapper
# ---------------------------------
def ardnmf(
    X: pd.DataFrame,
    K0: Union[int, None] = None,
    objective: str = 'poisson',
    max_iter: int = 10000,
    del_: int = 1,
    tolerance: float = 1e-6,
    phi: float = 1.0,
    a: float = 10.0,
    b: Union[float, None] = None ,
    prior_on_W: str = 'L1',
    prior_on_H: str = 'L1',
    report_freq: int = 100,
    active_thresh: float = 1e-2,
    cut_norm: float = 0.5,
    cut_diff: float = 1.0,
    cuda_int: Union[int, None] = 0,
    verbose: bool = True,
    tag: str = ""
    ) -> dict:
    """
    Wrapper for ARD-NMF. Wraps GPU implementaiton from:
    https://github.com/broadinstitute/SignatureAnalyzer-GPU
    ------------------------------------------------------------------------
    Args:
        * X: input matrix (features x samples)
        * K0: starting number of latent components
        * objective: objective function for optimizaiton
        * max_iter: maximum number of iterations for algorithm
        * del_: n/a
        * tolerance: stop point for optimization
        * phi: dispersion parameter
        * a: shape parameter
        * b: shape parameter
        * prior_on_W: L1 or L2
        * prior_on_H: L1 or L2
        * report_freq: how often to print stats
        * active_thresh: threshold for a latent component's impact on
            signature if the latent factor is less than this, it does not contribute
        * cut_norm: min normalized value for mean signature
            (used in post-processing)
        * cut_diff: difference between mean signature and rest of signatures
            for marker selction
            (used in post-processing)
        * cuda_int: GPU to use. Defaults to 0. If "None" or if no GPU available,
            will perform decomposition using CPU.
        * verbose: verbose reporting of algorithm convergence

    Returns:
        * results - dict with following keys:
            {'H', 'W', 'Wraw', 'Hraw', 'markers', 'signatures', 'objective', 'log', 'lam'}
    """
    assert objective in ('poisson','gaussian'), \
        "Unable to use {}; specify either poisson or gaussian objective.".format(objective)

    if objective == 'poisson': Beta = 1
    if objective == 'gaussian': Beta = 2

    assert prior_on_W in ('L1','L2'), \
        "Unable to use {}; use either L1 or L2 prior on W.".format(prior_on_W)
    assert prior_on_H in ('L1','L2'), \
        "Unable to use {}; use either L1 or L2 prior on H.".format(prior_on_H)

    # ---------------------------------
    # Load data into tensors
    # ---------------------------------
    data = ARD_NMF(X, objective, verbose=verbose)
    channel_names = data.channel_names
    sample_names = data.sample_names

    # ---------------------------------
    # Run NMF
    # ---------------------------------
    results = run_method_engine(
        data, \
        a, \
        phi, \
        b, \
        Beta, \
        prior_on_W, \
        prior_on_H, \
        K0, \
        tolerance, \
        max_iter, \
        report_freq=report_freq, \
        active_thresh=active_thresh, \
        cuda_int=cuda_int, \
        verbose=verbose, \
        tag=tag
    )

    W, H, nsig, nonzero_idx = transfer_weights(results[0], results[1], active_thresh=active_thresh)
    sig_names = [str(i) for i in range(1,nsig+1)]

    W = pd.DataFrame(data=W, index=channel_names, columns=sig_names)
    H = pd.DataFrame(data=H, index=sig_names, columns=sample_names)

    W,H = select_signatures(W,H)
    markers, signatures = select_markers(X, W, H, cut_norm=cut_norm, cut_diff=cut_diff, verbose=verbose)

    Wraw = pd.DataFrame(data=results[0][:,nonzero_idx], index=channel_names, columns=sig_names)
    Wraw = Wraw.rename(columns={x:'S'+x for x in Wraw.columns})
    Hraw = pd.DataFrame(data=results[1][nonzero_idx,:],  index=sig_names, columns=sample_names)
    Hraw = Hraw.rename(index={x:'S'+x for x in Hraw.index})

    # Fix log typing
    results[3]['K'] = results[3]['K'].astype(int)
    results[3]['obj'] = results[3]['obj'].astype('float')
    results[3]['b_div'] = results[3]['b_div'].astype('float')
    results[3]['lam'] = results[3]['lam'].astype('float')
    results[3]['del'] = results[3]['del'].astype('float')
    results[3]['W_sum'] = results[3]['W_sum'].astype('float')
    results[3]['H_sum'] = results[3]['H_sum'].astype('float')

    return {
        'H': H,
        'W': W,
        'Wraw':Wraw,
        'Hraw':Hraw,
        'markers': markers,
        'signatures': signatures,
        'objective': results[2],
        'log': results[3],
        'lam': results[4]
    }



import pandas as pd
import numpy as np
import sys
from sys import stdout
import argparse
import time
from scipy.special import gamma
import os
import pickle
import math
import torch
from typing import Union
import multiprocessing.connection as mpc
from .NMF_functions import *

class ARD_NMF:
    """
    NMF results class implements both half normal and exponential prior ARD NMF
    implementation based on https://arxiv.org/pdf/1111.6085.pdf
    """
    def __init__(self,dataset,objective,dtype = torch.float32, verbose=True):
        self.eps_ = torch.tensor(1.e-30,dtype=dtype,requires_grad=False)
        self.dataset = dataset
        zero_idx = np.sum(self.dataset, axis=1) > 0
        self.V0 = self.dataset.values[zero_idx, :]
        self.V = self.V0 - np.min(self.V0) + 1.e-30
        self.V_max = np.max(self.V)
        self.M = self.V.shape[0]
        self.N = self.V.shape[1]
        self.objective = objective
        self.channel_names = self.dataset.index[zero_idx]
        self.sample_names = self.dataset.columns
        self.dtype = dtype
        self.verbose = verbose
        if self.verbose: print('NMF class initalized.')

    def initalize_data(self,a,phi,b,prior_W,prior_H,Beta,K0,dtype = torch.float32):
        """
        Initializes dataset.

        Args:
            * a
            * phi: dispersion parameter - multiplied by variance if objective
                function is Gaussian (see Tan & Fevotte 2013)
            * b
            * prior_W
            * prior_H
            * Beta
            * K0: set to number of input features if not provided
        """
        if K0 == None:
            self.K0 = self.M
            self.number_of_active_components = self.M
        else:
            self.K0 = K0
            self.number_of_active_components = self.K0

        if self.objective.lower() == 'poisson':
            self.phi = torch.tensor(phi,dtype=dtype,requires_grad=False)
        else:
            self.phi = torch.tensor(np.var(self.V)* phi,dtype=dtype,requires_grad=False)

        self.a = a
        self.prior_W = prior_W
        self.prior_H = prior_H
        self.C = []
        self.b = b

        W0 = np.multiply(np.random.uniform(size=[self.M, self.K0])+self.eps_.numpy(), np.sqrt(self.V_max))
        H0 = np.multiply(np.random.uniform(size=[self.K0, self.N])+self.eps_.numpy(), np.sqrt(self.V_max))
        L0 = np.sum(W0,axis=0) + np.sum(H0,axis=1)

        self.W = torch.tensor(W0, dtype=self.dtype, requires_grad=False)
        self.H = torch.tensor(H0, dtype=self.dtype, requires_grad=False)
        self.Lambda = torch.tensor(L0, dtype=torch.float32, requires_grad=False)

        # calculate default b as described in Tan and Fevotte (2012)
        if self.b == None or self.b == 'None':
            # L1 ARD
            if self.prior_H == 'L1' and self.prior_W == 'L1':

                self.bcpu = np.sqrt(np.true_divide( (self.a - 1)*(self.a - 2) * np.mean(self.V),self.K0 ))
                self.b = torch.tensor(
                    np.sqrt(np.true_divide( (self.a - 1)*(self.a - 2) * np.mean(self.V),self.K0 ))
                    ,dtype=self.dtype,requires_grad=False)

                self.C = torch.tensor(self.N + self.M + self.a + 1, dtype=self.dtype, requires_grad=False)
            # L2 ARD
            elif self.prior_H == 'L2' and self.prior_W == 'L2':

                self.bcpu = np.true_divide(np.pi * (self.a - 1) * np.mean(self.V),2*self.K0)
                self.b = torch.tensor(
                    np.true_divide(np.pi * (self.a - 1) * np.mean(self.V),2*self.K0),
                    dtype=self.dtype,requires_grad=False)

                self.C = torch.tensor( (self.N + self.M)*0.5 + self.a + 1, dtype=self.dtype,requires_grad=False)

            # L1 - L2 ARD
            elif self.prior_H == 'L1' and self.prior_W == 'L2':
                self.bcpu = np.true_divide(np.mean(self.V)*np.sqrt(2)*gamma(self.a-3/2),self.K0*np.sqrt(np.pi)*gamma(self.a))
                self.b = torch.tensor(
                    np.true_divide(np.mean(self.V)*np.sqrt(2)*gamma(self.a-3/2),self.K0*np.sqrt(np.pi)*gamma(self.a))
                    ,dtype=self.dtype,requires_grad=False)
                self.C = torch.tensor(self.N + self.M/2 + self.a + 1, dtype=self.dtype)
            elif self.prior_H == 'L2' and self.prior_W == 'L1':
                self.bcpu = np.true_divide(np.mean(self.V)*np.sqrt(2)*gamma(self.a-3/2),self.K0*np.sqrt(np.pi)*gamma(self.a))
                self.b = torch.tensor(
                    np.true_divide(np.mean(self.V)*np.sqrt(2)*gamma(self.a-3/2),self.K0*np.sqrt(np.pi)*gamma(self.a)),
                    dtype=self.dtype,requires_grad=False)
                self.C = torch.tensor(self.N/2 + self.M + self.a + 1, dtype=self.dtype)
        else:
            self.bcpu = self.b
            self.b = torch.tensor(self.b, dtype=self.dtype,requires_grad=False)
            if self.prior_H == 'L1' and self.prior_W == 'L1':
                self.C = torch.tensor(self.N + self.M + self.a + 1, dtype=self.dtype,requires_grad=False)
            # L2 ARD
            elif self.prior_H == 'L2' and self.prior_W == 'L2':
                self.C = torch.tensor( (self.N + self.M)*0.5 + self.a + 1, dtype=self.dtype,requires_grad=False)
            # L1 - L2 ARD
            elif self.prior_H == 'L1' and self.prior_W == 'L2':
                self.C = torch.tensor(self.N + self.M/2 + self.a + 1, dtype=self.dtype,requires_grad=False)
            elif self.prior_H == 'L2' and self.prior_W == 'L1':
                self.C = torch.constant(self.N/2 + self.M + self.a + 1, dtype=self.dtype,requires_grad=False)

        self.V = torch.tensor(self.V,dtype=self.dtype,requires_grad=False)
        if self.verbose: print('NMF data and parameters set.')

    def get_number_of_active_components(self):
        self.number_of_active_components = torch.sum(torch.sum(self.W,0)> 0.0, dtype=self.dtype)

def print_report(iter,report,verbose,tag):
    """
    Prints report.
    """
    if verbose:
        print("nit={:>5} K={:>5} | obj={:.2f}\tb_div={:.2f}\tlam={:.2f}\tdel={:.8f}\tsumW={:.2f}\tsumH={:.2f}".format(
            iter,
            report[iter]['K'],
            report[iter]['obj'],
            report[iter]['b_div'],
            report[iter]['lam'],
            report[iter]['del'],
            report[iter]['W_sum'],
            report[iter]['H_sum']
            )
        )
    else:
        stdout.write("\r{}nit={:>5} K={} \tdel={:.8f}".format(
            tag,
            iter,
            report[iter]['K'],
            report[iter]['del']
            )
        )

def run_method_engine(
    results: ARD_NMF,
    a: float,
    phi: float,
    b: float,
    Beta: int,
    W_prior: str,
    H_prior: str,
    K0: int,
    tolerance: float,
    max_iter: int,
    report_freq: int = 10,
    active_thresh: float = 1e-5,
    send_end: Union[mpc.Connection, None] = None,
    cuda_int: Union[int, None] = 0,
    verbose: bool = True,
    tag: str = ""
    ) -> (pd.DataFrame, pd.DataFrame, np.ndarray, pd.DataFrame, np.ndarray):
    """
    Run ARD-NMF Engine.
    ------------------------------------------------------------------------
    Args:
        * results: initialized ARD_NMF class
        * a: shape parameter
        * phi: dispersion parameter
        * b: shape parameter
        * Beta: defined by objective function
        * W_prior: prior on W matrix ("L1" or "L2")
        * H_prior: prior on H matrix ("L1" or "L2")
        * K0: starting number of latent components
        * tolerance: end-point of optimization
        * max_iter: maximum number of iterations for algorithm
        * report_freq: how often to print updates
        * active_thresh: threshold for a latent component's impact on
            signature if the latent factor is less than this, it does not contribute
        * send_end: mpc.Connection resulting from multiprocessing.Pipe,
            for use in parameter sweep implementation
        * cuda_int: GPU to use. Defaults to 0. If "None" or if no GPU available,
            will perform decomposition using CPU.
        * verbose: verbose logging

    Returns:
        * H: (samples x K)
        * W: (K x features)
        * markers
        * signatures
    """
    # initalize the NMF run
    results.initalize_data(a, phi, b, W_prior, H_prior, Beta, K0)
    # specify GPU
    cuda_string = 'cuda:'+str(cuda_int)
    # copy data to GPU
    if torch.cuda.device_count() > 0 and cuda_int is not None:
        if verbose: print("   * Using GPU: {}".format(cuda_string))
        W,H,V,Lambda,C,b0,eps_,phi = results.W.cuda(cuda_string),results.H.cuda(cuda_string),results.V.cuda(cuda_string),results.Lambda.cuda(cuda_string),results.C.cuda(cuda_string),results.b.cuda(cuda_string),results.eps_.cuda(cuda_string),results.phi.cuda(cuda_string)
    else:
        W,H,V,Lambda,C,b0,eps_,phi = results.W,results.H,results.V,results.Lambda,results.C,results.b,results.eps_,results.phi
        if verbose: print("   * Using CPU")

    # tracking variables
    deltrack = 1000
    times = list()
    report = dict()
    iter = 0
    lam_previous = Lambda
    if verbose: print('%%%%%%%%%%%%%%%')
    if verbose: print('a =',results.a)
    if verbose: print('b =',results.bcpu)
    if verbose: print('%%%%%%%%%%%%%%%')

    # set method
    method = NMF_algorithim(Beta, H_prior, W_prior)

    while deltrack >= tolerance and iter < max_iter:
        # compute updates
        H,W,Lambda = method.forward(W,H,V,Lambda,C,b0,eps_,phi)

        # compute objective and cost
        l_ = beta_div(Beta,V,W,H,eps_)
        cost_ = calculate_objective_function(Beta,V,W,H,Lambda,C,eps_,phi,results.K0)

        # update tracking
        deltrack = torch.max(torch.div(torch.abs(Lambda-lam_previous), lam_previous+1e-30))
        lam_previous = Lambda

        # ---------------------------- Reporting ---------------------------- #
        if iter % report_freq == 0:
            report[iter] = {
                'K': torch.sum((torch.sum(H,1) * torch.sum(W,0))>active_thresh).cpu().numpy(),
                'obj': cost_.cpu().numpy(),
                'b_div': l_.cpu().numpy(),
                'lam': torch.sum(Lambda).cpu().numpy(),
                'del': deltrack.cpu().numpy(),
                'W_sum': torch.sum(W).cpu().numpy(),
                'H_sum': torch.sum(H).cpu().numpy()
            }
            print_report(iter,report,verbose,tag)
        # ------------------------------------------------------------------- #
        iter+=1


    # --------------------------- Final Report --------------------------- #
    report[iter] = {
        'K': torch.sum((torch.sum(H,1) * torch.sum(W,0))>active_thresh).cpu().numpy(),
        'obj': cost_.cpu().numpy(),
        'b_div': l_.cpu().numpy(),
        'lam': torch.sum(Lambda).cpu().numpy(),
        'del': deltrack.cpu().numpy(),
        'W_sum': torch.sum(W).cpu().numpy(),
        'H_sum': torch.sum(H).cpu().numpy()
    }
    print_report(iter,report,verbose,tag)

    if not verbose:
        stdout.write("\n")
    # ------------------------------------------------------------------- #

    if send_end:
        send_end.send([W.cpu().numpy(), H.cpu().numpy(), cost_.cpu().numpy()])
    else:
        final_report = pd.DataFrame.from_dict(report).T
        final_report.index.name = 'iter'
        return W.cpu().numpy(), H.cpu().numpy(), cost_.cpu().numpy(), final_report, Lambda.cpu().numpy()
